"""
Module for the FlowNode class.
"""

import mdp
import warnings as _warnings
import copy as _copy


class FlowNode(mdp.Node):
    """FlowNode wraps a Flow of Nodes into a single Node.

    This is handy if you want to use a flow where a Node is required.
    Additional args and kwargs for train and execute are supported.

    Note that for nodes in the internal flow the intermediate training phases
    will generally be closed, e.g. a CheckpointSaveFunction should not expect
    these training  phases to be left open.

    All the read-only container slots are supported and are forwarded to the
    internal flow.
    """

    def __init__(self, flow, input_dim=None, output_dim=None, dtype=None):
        """Wrap the given flow into this node.

        Pretrained nodes are allowed, but the internal flow should not
        be modified after the FlowNode was created (this will cause problems
        if the training phase structure of the internal nodes changes).

        If the node dimensions and dtype are not specified, they will be
        extracted from the internal nodes (late dimension setting is also
        supported).

        flow can have crash recovery enabled, but there is no special support
        for it.
        """
        self._flow = flow
        # set properties if needed:
        if input_dim is None:
            input_dim = self._flow[0].input_dim
        if output_dim is None:
            output_dim = self._flow[-1].output_dim
        if dtype is None:
            dtype = self._flow[-1].dtype
        # store which nodes are pretrained up to what phase
        self._pretrained_phase = [node.get_current_train_phase()
                                  for node in flow]
        # check if all the nodes are already fully trained
        train_len = 0
        for i_node, node in enumerate(self._flow):
            if node.is_trainable():
                train_len += (len(node._get_train_seq())
                              - self._pretrained_phase[i_node])
        if train_len:
            self._is_trainable = True
        else:
            self._is_trainable = False
        # remaining standard node initialisation
        super(FlowNode, self).__init__(input_dim=input_dim,
                                       output_dim=output_dim, dtype=dtype)

    @property
    def flow(self):
        """Read-only internal flow property.

        In general the internal flow should not be modified (see __init__
        for more details).
        """
        return self._flow

    def _set_input_dim(self, n):
        # try setting the input_dim of the first node
        self._flow[0].input_dim = n
        # let a consistency check run
        self._flow._check_nodes_consistency()
        # if we didn't fail here, go on
        self._input_dim = n

    def _set_output_dim(self, n):
        last_node = self._flow[-1]
        if len(self._flow) == 1:
            self._flow[-1].output_dim = n
        elif last_node.output_dim is None:
            self._fix_nodes_dimensions()
            # check if it worked
            if last_node.output_dim is None:
                if last_node.input_dim is None:
                    err = ("FlowNode can't set the dimension of the last "
                           "node, because its input_dim is undefined ("
                           "which could lead to inconsistent dimensions).")
                    raise mdp.InconsistentDimException(err)
                # now we can safely try to set the dimension
                last_node.output_dim = n
        # the last_node dim is now set 
        if n != last_node.output_dim:
            err = (("FlowNode can't be set to output_dim %d" % n) +
                   " because the last internal node already has " +
                   "output_dim %d." % last_node.output_dim)
            raise mdp.InconsistentDimException(err)
        self._output_dim = n
    
    def _fix_nodes_dimensions(self):
        """Try to fix the dimensions of the internal nodes."""
        if len(self._flow) > 1:
            prev_node = self._flow[0]
            for node in self._flow[1:]:
                if node.input_dim is None:
                    node.input_dim = prev_node.output_dim
                prev_node = node
            self._flow._check_nodes_consistency()
        if self._flow[-1].output_dim is not None:
            # additional checks are performed here
            self.output_dim = self._flow[-1].output_dim

    def _set_dtype(self, t):
        # dtype can not be set for sure in arbitrary flows
        # but here we want to be sure that FlowNode *can*
        # offer a dtype that is consistent
        for node in self._flow:
            node.dtype = t
        self._dtype = t

    def _get_supported_dtypes(self):
        # we support the minimal common dtype set
        types = set(mdp.utils.get_dtypes('All'))
        for node in self._flow:
            types = types.intersection(node.get_supported_dtypes())
        return list(types)

    def is_trainable(self):
        return self._is_trainable

    def is_invertible(self):
        return all(node.is_invertible() for node in self._flow)

    def _get_train_seq(self):
        """Return a training sequence containing all training phases."""
        
        def get_train_function(_i_node, _node):
            # This internal function is needed to channel the data through
            # the nodes in front of the current nodes.
            # using nested scopes here instead of default args, see pep-0227
            def _train(x, *args, **kwargs):
                if i_node > 0:
                    _node.train(self._flow.execute(x, nodenr=_i_node-1),
                                *args, **kwargs)
                else:
                    _node.train(x, *args, **kwargs)
            return _train
        
        train_seq = []
        for i_node, node in enumerate(self._flow):
            if node.is_trainable():
                remaining_len = (len(node._get_train_seq())
                                 - self._pretrained_phase[i_node])
                train_seq += ([(get_train_function(i_node, node),
                                node.stop_training)] * remaining_len)

        # try fix the dimension of the internal nodes and the FlowNode
        # after the last node has been trained
        def _get_stop_training_wrapper(self, node, func):
            def _stop_training_wrapper(*args, **kwargs):
                func(*args, **kwargs)
                self._fix_nodes_dimensions()
            return _stop_training_wrapper
        
        if train_seq:
            train_seq[-1] = (train_seq[-1][0],
                             _get_stop_training_wrapper(self, self._flow[-1],
                                                        train_seq[-1][1]))
        return train_seq

    def _execute(self, x, *args, **kwargs):
        return self._flow.execute(x, *args, **kwargs)

    def _inverse(self, x):
        return self._flow.inverse(x)

    def copy(self, protocol=None):
        """Return a copy of this node.

        The copy call is delegated to the internal node, which allows the use
        of custom copy methods for special nodes.

        The protocol parameter should not be used.
        """
        if protocol is not None:
            _warnings.warn("protocol parameter to copy() is ignored",
                           mdp.MDPDeprecationWarning, stacklevel=2)
        # Warning: If we create a new FlowNode with the copied internal
        #    nodes then it will differ from the original one if some nodes
        #    were trained in the meantime. Especially _get_train_seq would
        #    return a shorter list in that case, possibly breaking stuff
        #    outside of this FlowNode (e.g. if it is enclosed by another
        #    FlowNode the _train_phase of this node will no longer fit the
        #    result of _get_train_seq).
        #
        # copy the nodes by delegation
        old_nodes = self._flow[:]
        new_nodes = [node.copy() for node in old_nodes]
        # now copy the rest of this flownode via deepcopy
        self._flow.flow = None
        new_flownode = _copy.deepcopy(self)
        new_flownode._flow.flow = new_nodes
        self._flow.flow = old_nodes
        return new_flownode

    ## container methods ##

    def __len__(self):
        return len(self._flow)

    def __getitem__(self, key):
        return self._flow.__getitem__(key)

    def __contains__(self, item):
        return self._flow.__contains__(item)

    def __iter__(self):
        return self._flow.__iter__()



from builtins import range
import py.test
from ._tools import *

INDIM, OUTDIM, TLEN = 5, 3, 10000

def train_LRNode(inp, out, with_bias):
    lrnode = mdp.nodes.LinearRegressionNode(with_bias)
    for i in range(len(inp)):
        lrnode.train(inp[i], out[i])
    lrnode.stop_training()
    return lrnode

def test_LinearRegressionNode():
    # 1. first, without noise
    # 1a   without bias term
    # regression coefficients
    beta = numx_rand.uniform(-10., 10., size=(INDIM, OUTDIM))
    # input data
    x = numx_rand.uniform(-20., 20., size=(TLEN, INDIM))
    # output of the linear model
    y = mult(x, beta)
    # train
    lrnode = train_LRNode([x], [y], False)
    # test results
    assert_array_almost_equal(lrnode.beta, beta, decimal)
    res = lrnode(x)
    assert_array_almost_equal(res, y, decimal)

def test_LinearRegressionNode_with_bias():
    # 1b with bias
    beta = numx_rand.uniform(-10., 10., size=(INDIM+1, OUTDIM))
    x = numx_rand.uniform(-20., 20., size=(TLEN, INDIM))
    y = mult(x, beta[1:,:]) + beta[0,:]
    lrnode = train_LRNode([x], [y], True)
    assert_array_almost_equal(lrnode.beta, beta, decimal)
    res = lrnode(x)
    assert_array_almost_equal(res, y, decimal)

def test_LinearRegressionNode_with_noise():
    # 2. with noise, multiple sets of input
    beta = numx_rand.uniform(-10., 10., size=(INDIM+1, OUTDIM))
    x = numx_rand.uniform(-20., 20., size=(TLEN, INDIM))
    y = mult(x, beta[1:,:]) + beta[0,:]

    inp = [numx_rand.uniform(-20., 20., size=(TLEN, INDIM))
           for i in range(5)]
    out = [mult(x, beta[1:,:]) + beta[0,:] +
           numx_rand.normal(size=y.shape)*0.1 for x in inp]
    lrnode = train_LRNode(inp, out, True)
    assert_array_almost_equal(lrnode.beta, beta, 2)
    res = lrnode(inp[0])
    assert_array_almost_equal_diff(res, out[0], 2)

@skip_on_condition('sys.version_info.major == 3','This test always fails on Python3')
def test_LinearRegressionNode_raises_on_linearly_dependent_input():
    # 3. test error for linearly dependent input
    beta = numx_rand.uniform(-10., 10., size=(INDIM, OUTDIM))
    x = numx.linspace(-20,20,TLEN)
    x = mdp.utils.rrep(x, INDIM)
    x[:,-1] = 2.*x[:,0]
    y = mult(x, beta)
    py.test.raises(mdp.NodeException, train_LRNode, [x], [y], False)

def test_LinearRegressionNode_raises_on_wrong_output_size():
    # 4. test wrong output size
    beta = numx_rand.uniform(-10., 10., size=(INDIM, OUTDIM))
    x = numx_rand.uniform(-20., 20., size=(TLEN, INDIM))
    x[:,-1] = 2.*x[:,0]
    y = mult(x, beta)
    y = y[:10,:]
    py.test.raises(mdp.TrainingException, train_LRNode, [x], [y], False)



from builtins import range
from ._tools import *
from mdp.nodes import TimeDelayNode, TimeDelaySlidingWindowNode

def test_TimeDelayNodes():
    x = numx.array(
        [ [1, 2, 3] , 
          [2, 4, 4],  
          [3, 8, 5], 
          [4, 16, 6],
          [5, 32, 7] 
          ])

    ################################################
    node = TimeDelayNode(time_frames=3, gap=2)
    slider = TimeDelaySlidingWindowNode(time_frames=3, gap=2)

    real_res = [ [1, 2, 3, 0, 0, 0, 0, 0, 0],
                 [2, 4, 4, 0, 0, 0, 0, 0, 0],
                 [3, 8, 5, 1, 2, 3, 0, 0, 0],
                 [4, 16,6, 2, 4, 4, 0, 0, 0],
                 [5, 32,7, 3, 8, 5, 1, 2, 3]
                ]
    real_res = numx.array(real_res)

    res = node.execute(x)

    assert_array_equal(real_res, res)

    # test sliding window
    slider_res = numx.zeros_like(real_res)
    for row_nr in range(x.shape[0]):
        slider_res[row_nr, :] = slider.execute(x[[row_nr], :])

    assert_array_equal(real_res, slider_res)

    ################################################
    node = TimeDelayNode(time_frames=2, gap=3)
    slider = TimeDelaySlidingWindowNode(time_frames=2, gap=3)

    real_res = [ [1, 2, 3, 0, 0, 0],
                 [2, 4, 4, 0, 0, 0],
                 [3, 8, 5, 0, 0, 0],
                 [4, 16,6, 1, 2, 3],
                 [5, 32,7, 2, 4, 4]
                ]
    real_res = numx.array(real_res)

    res = node.execute(x)
    assert_array_equal(real_res, res)

    # test sliding window
    slider_res = numx.zeros_like(real_res)
    for row_nr in range(x.shape[0]):
        slider_res[row_nr, :] = slider.execute(x[[row_nr], :])

    assert_array_equal(real_res, slider_res)

    ################################################
    node = TimeDelayNode(time_frames=4, gap=1)
    slider = TimeDelaySlidingWindowNode(time_frames=4, gap=1)

    real_res = [ [1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                 [2, 4, 4, 1, 2, 3, 0, 0, 0, 0, 0, 0],
                 [3, 8, 5, 2, 4, 4, 1, 2, 3, 0, 0, 0],
                 [4, 16,6, 3, 8, 5, 2, 4, 4, 1, 2, 3],
                 [5, 32,7, 4, 16,6, 3, 8, 5, 2, 4, 4]
                ]
    real_res = numx.array(real_res)

    res = node.execute(x)
    assert_array_equal(real_res, res)

    # test sliding window
    slider_res = numx.zeros_like(real_res)
    for row_nr in range(x.shape[0]):
        slider_res[row_nr, :] = slider.execute(x[[row_nr], :])

    assert_array_equal(real_res, slider_res)




from mdp.test.test_namespace_fixups import (generate_calls,
                                            test_exports)

MODULES = ['bimdp',
           'bimdp.nodes',
           'bimdp.hinet',
           'bimdp.parallel',
           ]

def pytest_generate_tests(metafunc):
    generate_calls(MODULES, metafunc)




from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
__docformat__ = "restructuredtext en"

import mdp

class XSFANode(mdp.Node):
    """Perform Non-linear Blind Source Separation using Slow Feature Analysis.

    This node is designed to iteratively extract statistically
    independent sources from (in principle) arbitrary invertible
    nonlinear mixtures. The method relies on temporal correlations in
    the sources and consists of a combination of nonlinear SFA and a
    projection algorithm. More details can be found in the reference
    given below (once it's published).

    The node has multiple training phases. The number of training
    phases depends on the number of sources that must be
    extracted. The recommended way of training this node is through a
    container flow::

       >>> flow = mdp.Flow([XSFANode()])
       >>> flow.train(x)

    doing so will automatically train all training phases. The argument
    ``x`` to the ``Flow.train`` method can be an array or a list of iterables
    (see the section about Iterators in the MDP tutorial for more info).

    If the number of training samples is large, you may run into
    memory problems: use data iterators and chunk training to reduce
    memory usage.

    If you need to debug training and/or execution of this node, the
    suggested approach is to use the capabilities of BiMDP. For example::

       >>> flow = mdp.Flow([XSFANode()])
       >>> tr_filename = bimdp.show_training(flow=flow, data_iterators=x)
       >>> ex_filename, out = bimdp.show_execution(flow, x=x)

    this will run training and execution with bimdp inspection. Snapshots
    of the internal flow state for each training phase and execution step
    will be opened in a web brower and presented as a slideshow.

    References:
    Sprekeler, H., Zito, T., and Wiskott, L. (2009).
    An Extension of Slow Feature Analysis for Nonlinear Blind Source Separation.
    Journal of Machine Learning Research. 
    http://cogprints.org/7056/1/SprekelerZitoWiskott-Cogprints-2010.pdf
    """
    def __init__(self, basic_exp=None, intern_exp=None, svd=False, verbose=False,
                 input_dim=None, output_dim=None, dtype=None):
        """
        :Keywords:
           basic_exp
             a tuple ``(node, args, kwargs)`` defining the node used for
             the basic nonlinear expansion.  It is assumed that the
             mixture is linearly invertible after this expansion. The
             higher the complexity of the nonlinearity, the higher are
             the chances of inverting the unknown mixture. On the
             other hand, high complexity of the nonlinear expansion
             increases the danger of numeric instabilities, which can
             cause singularities in the simulation or errors in the
             source estimation. The trade-off has to be evaluated
             carefully.

             Default: ``(mdp.nodes.PolynomialExpansionNode, (2, ), {})``

           intern_exp
             a tuple ``(node, args, kwargs)`` defining the node used
             for the internal nonlinear expansion of the estimated
             sources to be removed from the input space.  The same
             trade-off as for basic_exp is valid here.

             Default: ``(mdp.nodes.PolynomialExpansionNode, (10, ), {})``

           svd
             enable Singular Value Decomposition for normalization
             and regularization. Use it if the node complains about
             singular covariance matrices.

           verbose
             show some progress during training.

             Default: False
        """

        # set up basic expansion
        if basic_exp is None:
            self.basic_exp = mdp.nodes.PolynomialExpansionNode
            self.basic_exp_args = (2, )
            self.basic_exp_kwargs = {}
        else:
            self.basic_exp = basic_exp[0]
            self.basic_exp_args = basic_exp[1]
            self.basic_exp_kwargs = basic_exp[2]

        # set up internal expansion
        if intern_exp is None:
            self.exp = mdp.nodes.PolynomialExpansionNode
            self.exp_args = (10, )
            self.exp_kwargs = {}
        else:
            self.exp = intern_exp[0]
            self.exp_args = intern_exp[1]
            self.exp_kwargs = intern_exp[2]

        # number of sources already extracted
        self.n_extracted_src = 0
        # internal network
        self._flow = None
        self.verbose = verbose
        self.svd = svd
        super(XSFANode, self).__init__(input_dim=input_dim,
                                       output_dim=output_dim, dtype=dtype)

    @property
    def flow(self):
        """Read-only internal flow property."""
        return self._flow

    def _get_train_seq(self):
        #XXX: this is a  hack
        # In order to enable the output_dim to be set automatically
        # after input_dim is known, instead of forcing the user to specify
        # it by hand, we need to initialize the internal flow just before
        # starting the first training (input_dim are known at that moment).
        # Problem is that when XSFANode is trained through a container flow,
        # which is the standard way of training this kind of nodes,
        # the flow checks that the data_iterators are *not* generators
        # for multiple phases nodes. To assess if a node has multiple phases
        # it checks that len(self._train_seq) > 1. But we still
        # don't know the number of training_phases at this point, because we
        # first need to know input_dim, which we will know after we receive the
        # first chunk of data. To avoid the flow to complain we just return
        # a bogus list of training phases: it should break anything else.
        if self._flow is None:
            # we still don't know the number of training_phases yet,
            # but we can assure that we will have more than 1:
            return [(None, None), (None, None)]
        else:
            return ([(self._train, self._stop_training)] *
                    sum(self._training_phases))

    def _set_input_dim(self, n):
        self._input_dim = n
        # set output_dim if thery are still not set
        if self.output_dim is None:
            self.output_dim = n

    def _check_train_args(self, x):
        # this method will be called before starting training.
        # it is the right moment to initialize the internal flow
        if self._flow is None:
            self._initialize_internal_flow()
            if self.verbose:
                print("Extracting source 1...")

    def _initialize_internal_flow(self):
        # create the initial flow if it's not there already
        # set input_dim is needed to correctly create the first
        # network layer
        self.basic_exp_kwargs['input_dim'] = self.input_dim
        exp = self.basic_exp(*self.basic_exp_args, **self.basic_exp_kwargs)
        # first element of the flow is the basic expansion node
        # after that the first source extractor module is appended
        self._flow = (exp + self._get_source_extractor(exp.output_dim, 0))

        # set the training phases
        # set the total number of training phases
        training_phases = []
        for S in range(self.output_dim):
            # get the number of training phases of every single
            # source extractor module
            mod = self._get_source_extractor(S+1, S)
            training_phases.append(len(mod._train_seq))

        self._training_phases = training_phases

        # this is a list of the training phases the correspond to
        # completed training of a source extractor module
        self._training_phases_mods = [sum(training_phases[:i+1]) for i in
                                      range(len(training_phases[:-1]))]

    @staticmethod
    def is_invertible():
        return False

    def _train(self, x):
        # train the last source extractor module in the flow
        self._flow[-1].train(self._flow[:-1](x))

    def _stop_training(self):
        # stop the current training phase
        self._flow[-1].stop_training()
        # update the current training phase
        cur_tr_ph = self.get_current_train_phase() + 1
        # if we finished to train the current source extractor module
        # and we still have to extract some sources
        # append a new source extractor module
        if (cur_tr_ph in self._training_phases_mods and
            self.n_extracted_src != (self.output_dim - 1)):

            self.n_extracted_src += 1
            mod = self._get_source_extractor(self._flow[-1].output_dim,
                                             self.n_extracted_src)
            self._flow.append(mod)
            if self.verbose:
                print("Extracting source %d..." % (self.n_extracted_src+1))

    def _execute(self, x):
        return self._flow(x)[:,:self.output_dim]

    def _get_source_extractor(self, dim, nsources):
        # returns a module to extract the next source and remove its
        # projections in the data space
        S = nsources
        L = dim-S

        # sfa - extracts the next source
        sfa = mdp.nodes.SFANode(input_dim=L, output_dim=L)

        # identity - copies the new sources
        idn_new1 = mdp.nodes.IdentityNode(input_dim=S+1)
        # source expansion
        self.exp_kwargs['input_dim'] = S + 1
        # N2
        src_exp = mdp.hinet.FlowNode(self.exp(*self.exp_args,
                                              **self.exp_kwargs) +
                                     NormalizeNode() +
                                     mdp.nodes.WhiteningNode(svd=self.svd,
                                                             reduce=True))
        N2Layer = mdp.hinet.SameInputLayer((src_exp, idn_new1))
        N2ContLayer = mdp.hinet.Layer((N2Layer,
                                       mdp.nodes.IdentityNode(input_dim=L-1)))

        if S == 0:
            # don't need to copy the current sources (there are none)
            N1 = mdp.hinet.FlowNode(sfa + N2ContLayer)
        elif S == self.output_dim - 1:
            # the last source does not need to be removed
            # take care of passing the sources down along the flow
            idn_old = mdp.nodes.IdentityNode(input_dim=S)
            return mdp.hinet.Layer((idn_old,
                                    mdp.nodes.SFANode(input_dim=L,
                                                      output_dim=1)))
        else:
            # take care of passing the sources down along the flow
            idn_old = mdp.nodes.IdentityNode(input_dim=S)
            N1 = mdp.hinet.FlowNode(mdp.hinet.Layer((idn_old, sfa)) +
                                    N2ContLayer)

        # expanded sources projection
        proj = ProjectionNode(S, L-1)
        # use another identity node to copy the sources
        # we could in principle reuse the idn_new1 but using a new
        # node will make debugging much easier
        idn_new2 = mdp.nodes.IdentityNode(input_dim=S+1)
        # regularization after projection + new source copying
        reg_and_copy = mdp.hinet.Layer((idn_new2,
                                        mdp.nodes.WhiteningNode(input_dim=L-1,
                                                                svd=self.svd,
                                                                reduce=True)))
        # actual source removal flow
        src_rem = mdp.hinet.FlowNode( proj + reg_and_copy )
        # return the actual source extraction module
        return mdp.hinet.FlowNode(N1 + src_rem)


class ProjectionNode(mdp.Node):
    """Get expanded sources and input signals, and return
    the sources and the input signals projected into the space
    orthogonal to the expanded sources and their products."""
    def __init__(self, S, L):
        #!! IMPORTANT!!
        # this node *must* return the sources together with the
        # projected input signals
        self.proj_mtx = None
        self.L = L
        super(ProjectionNode, self).__init__(output_dim=S+1+L)
        self._cov_mtx = mdp.utils.CrossCovarianceMatrix(self.dtype)

    def _train(self, x):
        # compute covariance between expanded sources
        # and input signals
        self._cov_mtx.update(x[:,:-self.output_dim], x[:,-self.L:])

    def _stop_training(self):
        self.proj_mtx, avgx, avgy, self.tlen = self._cov_mtx.fix()

    def _execute(self, x):
        src = x[:, -self.output_dim:-self.L]
        exp = x[:, :-self.output_dim]
        inp = x[:, -self.L:]
        # result container
        result = mdp.numx.zeros((x.shape[0], self.output_dim))
        # project input on the plane orthogonal to the expanded sources
        result[:, -self.L:] = inp - mdp.utils.mult(exp, self.proj_mtx)
        # copy the sources
        result[:, :-self.L] = src
        return result

class NormalizeNode(mdp.PreserveDimNode):
    """Make input signal meanfree and unit variance"""
    def __init__(self, input_dim=None, output_dim=None, dtype=None):
        self._cov_mtx = mdp.utils.CovarianceMatrix(dtype)
        super(NormalizeNode, self).__init__(input_dim, output_dim, dtype)

    @staticmethod
    def is_trainable():
        return True

    def _train(self, x):
        self._cov_mtx.update(x)

    def _stop_training(self):
        cov_mtx, avg, tlen = self._cov_mtx.fix()
        self.m = avg
        self.s = mdp.numx.sqrt(mdp.numx.diag(cov_mtx))

    def _execute(self, x):
        return old_div((x - self.m),self.s)

    def _inverse(self, y):
        return y*self.s + self.m



from __future__ import division
from builtins import zip
from builtins import range
from builtins import object
from past.utils import old_div
from ._tools import *

def _randomly_filled_hypercube(widths, num_elem=1000):
    """Fills a hypercube with given widths, centred at the origin.
    """
    p = []
    for i in range(num_elem):
        rand_data = numx_rand.random(len(widths))
        rand_data = [w*(d - 0.5) for d, w in zip(rand_data, widths)]
        p.append(tuple(rand_data))
    return p

def _randomly_filled_hyperball(dim, radius, num_elem=1000):
    """Fills a hyperball with a number of random elements.
    """
    r = numx_rand.random(num_elem)
    points = numx_rand.random((num_elem, dim))
    for i in range(len(points)):
        norm = numx.linalg.norm(points[i])
        scale = pow(r[i], old_div(1.,dim))
        points[i] = points[i] * radius * scale / norm
    return points

def _random_clusters(positions, radius=1, num_elem=1000):
    """Puts random clusters with num_elem elements at the given positions.
    positions - a list of tuples
    """
    data = []
    for p in positions:
        dim = len(p)
        ball = _randomly_filled_hyperball(dim, radius, num_elem)
        ball = [numx.array(b) + numx.array(p) for b in ball]
        data.append(ball)
    return data

def _separable_data(positions, labels, radius=1, num_elem=1000, shuffled=False):
    """
    For each position, we create num_elem data points in a certain radius around
    that position. If shuffled, we shuffle the output data and labels.
    
    positions -- List of position tuples, e.g. [(1, 1), (-1, -1)]
    labels    -- List of labels, e.g. [1, -1]
    radius    -- The maximum distance to the position
    num_elem  -- The number of elements to be created
    shuffled  -- Should the output be shuffled.
    
    Returns:
      data, labels
    """
    assert len(positions) == len(labels)

    data = numx.vstack( _random_clusters(positions, radius, num_elem) )
    #data = numx.vstack( (numx_rand.random( (num_elem,2) ) - dist,
    #                     numx_rand.random( (num_elem,2) ) + dist) )
    a_labels = numx.hstack([[x] * num_elem for x in labels])
    if shuffled:
        ind = list(range(len(data)))
        numx_rand.shuffle(ind)
        return data[ind], a_labels[ind]
    return data, a_labels

def _sqdist(tuple_a, tuple_b):
    return sum( (a-b)**2 for a, b in zip(tuple_a, tuple_b) )

def test_separable_data_is_inside_radius():
    positions = [[(1, 1), (-1, -1)],
                 [(1, 1, 10), (100, -20, 30), (-1, 10, 1000)]]
    labels = [[1, -1], [1, 2, 3]]
    radii = [0.5, 1, 10]
    num_elem = 100

    for pos, labs in zip(positions, labels):
        for rad in radii:
            data, ls = _separable_data(pos, labs, rad, num_elem)

            for d,l in zip(data, ls):
                idx = labs.index(l)
                assert rad**2 > _sqdist(pos[idx], d)

@skip_on_condition(
    "not hasattr(mdp.nodes, 'ShogunSVMClassifier')",
    "This test requires the 'shogun' module.")
def test_ShogunSVMClassifier():
    # TODO: Implement parameter ranges
    num_train = 100
    num_test = 50
    for positions in [((1,), (-1,)),
                      ((1,1), (-1,-1)),
                      ((1,1,1), (-1,-1,1)),
                      ((1,1,1,1), (-1,1,1,1)),
                      ((1,1,1,1), (-1,-1,-1,-1)),
                      ((1,1), (-1,-1), (1, -1), (-1, 1))
                      ]:

        radius = 0.3

        if len(positions) == 2:
            labels = (-1, 1)
        elif len(positions) == 3:
            labels = (-1, 1, 1)
        elif len(positions) == 4:
            labels = (-1, -1, 1, 1)

        traindata_real, trainlab = _separable_data(positions, labels,
                                                          radius, num_train)
        testdata_real, testlab = _separable_data(positions, labels,
                                                        radius, num_test)


        classifiers = ['GMNPSVM', 'GNPPSVM', 'GPBTSVM', #'KernelPerceptron',
                       'LDA', 'LibSVM', #'LibSVMOneClass', 'MPDSVM',
                       'Perceptron', 'SVMLin']
        kernels = ['PolyKernel', 'LinearKernel', 'SigmoidKernel', 'GaussianKernel']

        #kernels = list(mdp.nodes.ShogunSVMClassifier.kernel_parameters.keys())
        combinations = {'classifier': classifiers,
                        'kernel': kernels}

        for comb in utils.orthogonal_permutations(combinations):
            # this is redundant but makes it clear,
            # what has been taken out deliberately
            if comb['kernel'] in ['PyramidChi2', 'Chi2Kernel']:
                # We don't have good init arguments for these
                continue
            if comb['classifier'] in ['LaRank', 'LibLinear', 'LibSVMMultiClass',
                                      'MKLClassification', 'MKLMultiClass',
                                      'MKLOneClass', 'MultiClassSVM', 'SVM',
                                      'SVMOcas', 'SVMSGD', 'ScatterSVM',
                                      'SubGradientSVM']:
                # We don't have good init arguments for these and/or they work differently
                continue

            # something does not work here: skipping
            if comb['classifier'] == 'GPBTSVM' and comb['kernel'] == 'LinearKernel':
                continue

            sg_node = mdp.nodes.ShogunSVMClassifier(classifier=comb['classifier'])

            if sg_node.classifier.takes_kernel:
                sg_node.set_kernel(comb['kernel'])

            # train in two chunks to check update mechanism
            sg_node.train( traindata_real[:num_train], trainlab[:num_train] )
            sg_node.train( traindata_real[num_train:], trainlab[num_train:] )

            assert sg_node.input_dim == len(traindata_real.T)

            out = sg_node.label(testdata_real)

            if sg_node.classifier.takes_kernel:
                # check that the kernel has stored all our training vectors
                assert sg_node.classifier.kernel.get_num_vec_lhs() == num_train * len(positions)
                # check that the kernel has also stored the latest classification vectors in rhs
                assert sg_node.classifier.kernel.get_num_vec_rhs() == num_test * len(positions)

            # Test also for inverse
            worked = numx.all(numx.sign(out) == testlab) or \
                     numx.all(numx.sign(out) == -testlab)
            failed = not worked

            should_fail = False
            if len(positions) == 2:
                if comb['classifier'] in ['LibSVMOneClass',
                                          'GMNPSVM']:
                    should_fail = True
                if comb['classifier'] == 'GPBTSVM' and \
                   comb['kernel'] in ['LinearKernel']:
                    should_fail = True

            # xor problem
            if len(positions) == 4:
                if comb['classifier'] in ['LibSVMOneClass', 'SVMLin', 'Perceptron',
                                          'LDA', 'GMNPSVM']:
                    should_fail = True
                if comb['classifier'] == 'LibSVM' and \
                   comb['kernel'] in ['LinearKernel', 'SigmoidKernel']:
                    should_fail = True
                if comb['classifier'] == 'GPBTSVM' and \
                   comb['kernel'] in ['LinearKernel', 'SigmoidKernel']:
                    should_fail = True
                if comb['classifier'] == 'GNPPSVM' and \
                   comb['kernel'] in ['LinearKernel', 'SigmoidKernel']:
                    should_fail = True

            if should_fail:
                msg = ("Classification should fail but did not in %s. Positions %s." %
                      (sg_node.classifier, positions))
            else:
                msg = ("Classification should not fail but failed in %s. Positions %s." %
                      (sg_node.classifier, positions))

            assert should_fail == failed, msg



class TestLibSVMClassifier(object):

    @skip_on_condition("not hasattr(mdp.nodes, 'LibSVMClassifier')",
                       "This test requires the 'libsvm' module.")
    def setup_method(self, method):
        self.combinations = {'kernel': mdp.nodes.LibSVMClassifier.kernels,
                             'classifier': mdp.nodes.LibSVMClassifier.classifiers}

    def test_that_parameters_are_correct(self):
        import svm as libsvm
        for comb in utils.orthogonal_permutations(self.combinations):
            C = 1.01
            epsilon = 1.1e-5
            svm_node = mdp.nodes.LibSVMClassifier(params={"C": C, "eps": epsilon})
            svm_node.set_kernel(comb['kernel'])
            svm_node.set_classifier(comb['classifier'])
            
            # check that the parameters are correct
            assert svm_node.parameter.kernel_type == getattr(libsvm, comb['kernel'])
            assert svm_node.parameter.svm_type == getattr(libsvm, comb['classifier'])
            assert svm_node.parameter.C == C
            assert svm_node.parameter.eps == epsilon

    def test_linear_separable_data(self):
        num_train = 100
        num_test = 50
        C = 1.01
        epsilon = 1e-5
        for positions in [((1,), (-1,)),
                          ((1,1), (-1,-1)),
                          ((1,1,1), (-1,-1,1)),
                          ((1,1,1,1), (-1,1,1,1)),
                          ((1,1,1,1), (-1,-1,-1,-1))]:
            radius = 0.3

            traindata_real, trainlab = _separable_data(positions, (-1, 1),
                                                       radius, num_train, True)
            testdata_real, testlab = _separable_data(positions, (-1, 1),
                                                     radius, num_test, True)

            for comb in utils.orthogonal_permutations(self.combinations):
                # Take out non-working cases
                if comb['classifier'] in ["ONE_CLASS"]:
                    continue
                if comb['kernel'] in ["SIGMOID", "POLY"]:
                    continue
                if len(positions[0]) == 1 and comb['kernel'] == "RBF":
                    # RBF won't work in 1d
                    continue

                svm_node = mdp.nodes.LibSVMClassifier(kernel=comb['kernel'],
                                                      classifier=comb['classifier'],
                                                      probability=True,
                                                      params={"C": C, "eps": epsilon})
                
                # train in two chunks to check update mechanism
                svm_node.train(traindata_real[:num_train], trainlab[:num_train])
                svm_node.train(traindata_real[num_train:], trainlab[num_train:])

                assert svm_node.input_dim == len(traindata_real.T)

                out = svm_node.label(testdata_real)

                testerr = numx.all(numx.sign(out) == testlab)
                assert testerr, ('classification error for ', comb)

                # we don't have ranks in our regression models
                if not comb['classifier'].endswith("SVR"):
                    pos1_rank = numx.array(svm_node.rank(numx.array([positions[0]])))
                    pos2_rank = numx.array(svm_node.rank(numx.array([positions[1]])))

                    assert numx.all(pos1_rank == -pos2_rank)
                    assert numx.all(abs(pos1_rank) == 1)
                    assert numx.all(abs(pos2_rank) == 1)





from __future__ import division
from past.utils import old_div
from ._tools import *
mult = mdp.utils.mult

def testSFANode():
    dim=10000
    freqs = [2*numx.pi*1, 2*numx.pi*5]
    t =  numx.linspace(0,1,num=dim)
    mat = numx.array([numx.sin(freqs[0]*t), numx.sin(freqs[1]*t)]).T
    mat = (old_div((mat - mean(mat[:-1,:], axis=0)), std(mat[:-1,:],axis=0)))
    des_mat = mat.copy()
    mat = mult(mat,uniform((2,2))) + uniform(2)
    sfa = mdp.nodes.SFANode()
    sfa.train(mat)
    out = sfa.execute(mat)
    correlation = old_div(mult(des_mat[:-1,:].T,out[:-1,:]),(dim - 2))
    assert sfa.get_eta_values(t=0.5) is not None, 'get_eta is None'
    assert_array_almost_equal(abs(correlation),
                              numx.eye(2), decimal-3)
    sfa = mdp.nodes.SFANode(output_dim = 1)
    sfa.train(mat)
    out = sfa.execute(mat)
    assert out.shape[1]==1, 'Wrong output_dim'
    correlation = old_div(mult(des_mat[:-1,:1].T,out[:-1,:]),(dim - 2))
    assert_array_almost_equal(abs(correlation),
                              numx.eye(1), decimal - 3)

def testSFANode_range_argument():
    node = mdp.nodes.SFANode()
    x = numx.random.random((100,10))
    node.train(x)
    node.stop_training()
    y = node.execute(x, n=5)
    assert y.shape[1] == 5

def testSFANode_one_time_samples():
    # when training with x.shape = (1, n), stop_training
    # was failing with a ValueError: array must not contain infs or NaNs
    # because with only one samples no time difference can be computed and
    # the covmatrix is updated with zeros!
    node = mdp.nodes.SFANode()
    x = numx.random.random((1,5))
    with py.test.raises(mdp.TrainingException):
        node.train(x)

def testSFANode_include_last_sample():
    # check that the default behaviour is True
    node = mdp.nodes.SFANode()
    x = numx.random.random((100,10))
    node.train(x)
    node.stop_training()
    assert node.tlen == 100
    assert node.dtlen == 99

    # check that you can set it explicitly
    node = mdp.nodes.SFANode(include_last_sample=True)
    x = numx.random.random((100,10))
    node.train(x)
    node.stop_training()
    assert node.tlen == 100
    assert node.dtlen == 99

    # check the old behaviour
    node = mdp.nodes.SFANode(include_last_sample=False)
    x = numx.random.random((100,10))
    node.train(x)
    node.stop_training()
    assert node.tlen == 99
    assert node.dtlen == 99

    # check that we can change it during training
    node = mdp.nodes.SFANode(include_last_sample=False)
    x = numx.random.random((100,10))
    node.train(x, include_last_sample=True)
    node.stop_training()
    assert node.tlen == 100
    assert node.dtlen == 99
    
def testSFANode_derivative_bug1D():
    # one dimensional worst case scenario
    T = 100
    x = numx.zeros((T,1))
    x[0,:] =  -1.
    x[-1,:] = +1.
    x /= x.std(ddof=1)
    sfa = mdp.nodes.SFANode(include_last_sample=True)
    sfa.train(x)
    sfa.stop_training(debug=True)
    xdot = sfa.time_derivative(x)
    tlen = xdot.shape[0]
    correct_dcov_mtx = old_div((xdot*xdot).sum(),(tlen-1))
    sfa_dcov_mtx = sfa.dcov_mtx
    # quantify the error
    error = abs(correct_dcov_mtx-sfa_dcov_mtx)[0,0]
    assert error < 10**(-decimal)
    # the bug was that we were calculating the covariance matrix
    # of the derivative, i.e.
    # sfa_dcov-mtx = (xdot*xdot).sum()/(tlen-1) - xdot.sum()**2/(tlen*(tlen-1))
    # so that the error in the estimated matrix was exactly
    # xdot.sum()**2/(tlen*(tlen-1))

def testSFANode_derivative_bug2D():
    T = 100
    x = numx.zeros((T,2))
    x[0,0] =  -1.
    x[-1,0] = +1.
    x[:,1] = numx.arange(T)
    x -= x.mean(axis=0)
    x /= x.std(ddof=1, axis=0)    
    sfa = mdp.nodes.SFANode(include_last_sample=True)
    sfa.train(x)
    sfa.stop_training(debug=True)
    xdot = sfa.time_derivative(x)
    tlen = xdot.shape[0]
    correct_dcov_mtx = old_div(mdp.utils.mult(xdot.T, xdot),(tlen-1))
    sfa_dcov_mtx = sfa.dcov_mtx
    # the bug was that we were calculating the covariance matrix
    # of the derivative, i.e.
    # sfa_dcov_mtx = mdp.utils.mult(xdot.T, xdot)/(tlen-1) - \
    #                numx.outer(xdot.sum(axis=0),
    #                           xdot.sum(axis=0))/(tlen*(tlen-1)))
    # so that the error in the estimated matrix was exactly
    # numx.outer(xdot.sum(axis=0),xdot.sum(axis=0))/(tlen*(tlen-1))
    error = abs(correct_dcov_mtx-sfa_dcov_mtx)
    assert_array_almost_equal(numx.zeros(error.shape), error, decimal)

def testSFANode_derivative_bug2D_eigen():
    # this is a copy of the previous test, where we
    # quantify the error in the estimated eigenvalues
    # and eigenvectors
    T = 100
    x = numx.zeros((T,2))
    x[0,0] =  -1.
    x[-1,0] = +1.
    x[:,1] = numx.arange(T)
    x -= x.mean(axis=0)
    x /= x.std(ddof=1, axis=0)    
    sfa = mdp.nodes.SFANode(include_last_sample=True)
    sfa.train(x)
    sfa.stop_training(debug=True)
    xdot = sfa.time_derivative(x)
    tlen = xdot.shape[0]
    correct_dcov_mtx = old_div(mdp.utils.mult(xdot.T, xdot),(tlen-1))
    eigvalues, eigvectors = sfa._symeig(correct_dcov_mtx,
                                        sfa.cov_mtx,
                                        range=None,
                                        overwrite=False)
    assert_array_almost_equal(eigvalues, sfa.d, decimal)
    assert_array_almost_equal(eigvectors, sfa.sf, decimal)



from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
__docformat__ = "restructuredtext en"

from mdp import numx, numx_linalg, Cumulator, TrainingException, MDPWarning
from mdp.utils import mult, nongeneral_svd, svd, sqrtm, symeig
import warnings as _warnings

# some useful functions
sqrt = numx.sqrt

# search XXX for locations where future work is needed

#########################################################
#  Locally Linear Embedding
#########################################################

class LLENode(Cumulator):
    """Perform a Locally Linear Embedding analysis on the data.

    **Internal variables of interest**

      ``self.training_projection``
          The LLE projection of the training data (defined when
          training finishes).

      ``self.desired_variance``
          variance limit used to compute intrinsic dimensionality.

    Based on the algorithm outlined in *An Introduction to Locally
    Linear Embedding* by L. Saul and S. Roweis, using improvements
    suggested in *Locally Linear Embedding for Classification* by
    D. deRidder and R.P.W. Duin.

    References: Roweis, S. and Saul, L., Nonlinear dimensionality
    reduction by locally linear embedding, Science 290 (5500), pp.
    2323-2326, 2000.

    Original code contributed by: Jake VanderPlas, University of Washington,
    """

    def __init__(self, k, r=0.001, svd=False, verbose=False,
                 input_dim=None, output_dim=None, dtype=None):
        """
        :Arguments:
           k
             number of nearest neighbors to use
           r
             regularization constant; if ``None``, ``r`` is automatically
             computed using the method presented in deRidder and Duin;
             this method involves solving an eigenvalue problem for
             every data point, and can slow down the algorithm
             If specified, it multiplies the trace of the local covariance
             matrix of the distances, as in Saul & Roweis (faster)
           svd
             if true, use SVD to compute the projection matrix;
             SVD is slower but more stable
           verbose
             if true, displays information about the progress
             of the algorithm
           output_dim
             number of dimensions to output or a float between 0.0 and
             1.0. In the latter case, ``output_dim`` specifies the desired
             fraction of variance to be explained, and the final
             number of output dimensions is known at the end of
             training (e.g., for ``output_dim=0.95`` the algorithm will
             keep as many dimensions as necessary in order to explain
             95% of the input variance)
        """

        if isinstance(output_dim, float) and output_dim <= 1:
            self.desired_variance = output_dim
            output_dim = None
        else:
            self.desired_variance = None

        super(LLENode, self).__init__(input_dim, output_dim, dtype)

        self.k = k
        self.r = r
        self.svd = svd
        self.verbose = verbose

    def _stop_training(self):
        Cumulator._stop_training(self)

        if self.verbose:
            msg = ('training LLE on %i points'
                   ' in %i dimensions...' % (self.data.shape[0],
                                             self.data.shape[1]))
            print(msg)

        # some useful quantities
        M = self.data
        N = M.shape[0]
        k = self.k
        r = self.r

        # indices of diagonal elements
        W_diag_idx = numx.arange(N)
        Q_diag_idx = numx.arange(k)

        if k > N:
            err = ('k=%i must be less than or '
                   'equal to number of training points N=%i' % (k, N))
            raise TrainingException(err)

        # determines number of output dimensions: if desired_variance
        # is specified, we need to learn it from the data. Otherwise,
        # it's easy
        learn_outdim = False
        if self.output_dim is None:
            if self.desired_variance is None:
                self.output_dim = self.input_dim
            else:
                learn_outdim = True

        # do we need to automatically determine the regularization term?
        auto_reg = r is None

        # determine number of output dims, precalculate useful stuff
        if learn_outdim:
            Qs, sig2s, nbrss = self._adjust_output_dim()

        # build the weight matrix
        #XXX future work:
        #XXX   for faster implementation, W should be a sparse matrix
        W = numx.zeros((N, N), dtype=self.dtype)

        if self.verbose:
            print(' - constructing [%i x %i] weight matrix...' % W.shape)

        for row in range(N):
            if learn_outdim:
                Q = Qs[row, :, :]
                nbrs = nbrss[row, :]
            else:
                # -----------------------------------------------
                #  find k nearest neighbors
                # -----------------------------------------------
                M_Mi = M-M[row]
                nbrs = numx.argsort((M_Mi**2).sum(1))[1:k+1]
                M_Mi = M_Mi[nbrs]
                # compute covariance matrix of distances
                Q = mult(M_Mi, M_Mi.T)

            # -----------------------------------------------
            #  compute weight vector based on neighbors
            # -----------------------------------------------

            #Covariance matrix may be nearly singular:
            # add a diagonal correction to prevent numerical errors
            if auto_reg:
                # automatic mode: correction is equal to the sum of
                # the (d_in-d_out) unused variances (as in deRidder &
                # Duin)
                if learn_outdim:
                    sig2 = sig2s[row, :]
                else:
                    sig2 = svd(M_Mi, compute_uv=0)**2
                r = numx.sum(sig2[self.output_dim:])
                Q[Q_diag_idx, Q_diag_idx] += r
            else:
                # Roweis et al instead use "a correction that
                #   is small compared to the trace" e.g.:
                # r = 0.001 * float(Q.trace())
                # this is equivalent to assuming 0.1% of the variance is unused
                Q[Q_diag_idx, Q_diag_idx] += r*Q.trace()

            #solve for weight
            # weight is w such that sum(Q_ij * w_j) = 1 for all i
            # XXX refcast is due to numpy bug: floats become double
            w = self._refcast(numx_linalg.solve(Q, numx.ones(k)))
            w /= w.sum()

            #update row of the weight matrix
            W[nbrs, row] = w

        if self.verbose:
            msg = (' - finding [%i x %i] null space of weight matrix\n'
                   '     (may take a while)...' % (self.output_dim, N))
            print(msg)

        self.W = W.copy()
        #to find the null space, we need the bottom d+1
        #  eigenvectors of (W-I).T*(W-I)
        #Compute this using the svd of (W-I):
        W[W_diag_idx, W_diag_idx] -= 1.

        #XXX future work:
        #XXX  use of upcoming ARPACK interface for bottom few eigenvectors
        #XXX   of a sparse matrix will significantly increase the speed
        #XXX   of the next step
        if self.svd:
            sig, U = nongeneral_svd(W.T, range=(2, self.output_dim+1))
        else:
            # the following code does the same computation, but uses
            # symeig, which computes only the required eigenvectors, and
            # is much faster. However, it could also be more unstable...
            WW = mult(W, W.T)
            # regularizes the eigenvalues, does not change the eigenvectors:
            WW[W_diag_idx, W_diag_idx] += 0.1
            sig, U = symeig(WW, range=(2, self.output_dim+1), overwrite=True)

        self.training_projection = U

    def _adjust_output_dim(self):
        # this function is called if we need to compute the number of
        # output dimensions automatically; some quantities that are
        # useful later are pre-calculated to spare precious time

        if self.verbose:
            print(' - adjusting output dim:')

        #otherwise, we need to compute output_dim
        #                  from desired_variance
        M = self.data
        k = self.k
        N, d_in = M.shape

        m_est_array = []
        Qs = numx.zeros((N, k, k))
        sig2s = numx.zeros((N, d_in))
        nbrss = numx.zeros((N, k), dtype='i')

        for row in range(N):
            #-----------------------------------------------
            #  find k nearest neighbors
            #-----------------------------------------------
            M_Mi = M-M[row]
            nbrs = numx.argsort((M_Mi**2).sum(1))[1:k+1]
            M_Mi = M_Mi[nbrs]
            # compute covariance matrix of distances
            Qs[row, :, :] = mult(M_Mi, M_Mi.T)
            nbrss[row, :] = nbrs

            #-----------------------------------------------
            # singular values of M_Mi give the variance:
            #   use this to compute intrinsic dimensionality
            #   at this point
            #-----------------------------------------------
            sig2 = (svd(M_Mi, compute_uv=0))**2
            sig2s[row, :sig2.shape[0]] = sig2

            #-----------------------------------------------
            # use sig2 to compute intrinsic dimensionality of the
            #   data at this neighborhood.  The dimensionality is the
            #   number of eigenvalues needed to sum to the total
            #   desired variance
            #-----------------------------------------------
            sig2 /= sig2.sum()
            S = sig2.cumsum()
            m_est = S.searchsorted(self.desired_variance)
            if m_est > 0:
                m_est += old_div((self.desired_variance-S[m_est-1]),sig2[m_est])
            else:
                m_est = old_div(self.desired_variance,sig2[m_est])
            m_est_array.append(m_est)

        m_est_array = numx.asarray(m_est_array)
        self.output_dim = int( numx.ceil( numx.median(m_est_array) ) )
        if self.verbose:
            msg = ('      output_dim = %i'
                   ' for variance of %.2f' % (self.output_dim,
                                              self.desired_variance))
            print(msg)

        return Qs, sig2s, nbrss

    def _execute(self, x):
        #----------------------------------------------------
        # similar algorithm to that within self.stop_training()
        #  refer there for notes & comments on code
        #----------------------------------------------------
        N = self.data.shape[0]
        Nx = x.shape[0]
        W = numx.zeros((Nx, N), dtype=self.dtype)

        k, r = self.k, self.r
        d_out = self.output_dim
        Q_diag_idx = numx.arange(k)

        for row in range(Nx):
            #find nearest neighbors of x in M
            M_xi = self.data-x[row]
            nbrs = numx.argsort( (M_xi**2).sum(1) )[:k]
            M_xi = M_xi[nbrs]

            #find corrected covariance matrix Q
            Q = mult(M_xi, M_xi.T)
            if r is None and k > d_out:
                sig2 = (svd(M_xi, compute_uv=0))**2
                r = numx.sum(sig2[d_out:])
                Q[Q_diag_idx, Q_diag_idx] += r
            if r is not None:
                Q[Q_diag_idx, Q_diag_idx] += r

            #solve for weights
            w = self._refcast(numx_linalg.solve(Q , numx.ones(k)))
            w /= w.sum()
            W[row, nbrs] = w

        #multiply weights by result of SVD from training
        return numx.dot(W, self.training_projection)

    @staticmethod
    def is_trainable():
        return True

    @staticmethod
    def is_invertible():
        return False


#########################################################
#  Hessian LLE
#########################################################


# Modified Gram-Schmidt
def _mgs(a):
    m, n = a.shape
    v = a.copy()
    r = numx.zeros((n, n))
    for i in range(n):
        r[i, i] = numx_linalg.norm(v[:, i])
        v[:, i] = old_div(v[:, i],r[i, i])
        for j in range(i+1, n):
            r[i, j] = mult(v[:, i], v[:, j])
            v[:, j] = v[:, j] - r[i, j]*v[:, i]
    # q is v
    return v, r

class HLLENode(LLENode):
    """Perform a Hessian Locally Linear Embedding analysis on the data.

    **Internal variables of interest**

      ``self.training_projection``
          the HLLE projection of the training data (defined when training
          finishes)

      ``self.desired_variance``
          variance limit used to compute intrinsic dimensionality.

    Implementation based on algorithm outlined in
    Donoho, D. L., and Grimes, C., Hessian Eigenmaps: new locally linear
    embedding techniques for high-dimensional data, Proceedings of the
    National Academy of Sciences 100(10): 5591-5596, 2003.

    Original code contributed by: Jake Vanderplas, University of Washington
    """

    #----------------------------------------------------
    # Note that many methods ar inherited from LLENode,
    #  including _execute(), _adjust_output_dim(), etc.
    # The main advantage of the Hessian estimator is to
    #  limit distortions of the input manifold.  Once
    #  the model has been trained, it is sufficient (and
    #  much less computationally intensive) to determine
    #  projections for new points using the LLE framework.
    #----------------------------------------------------

    def __init__(self, k, r=0.001, svd=False, verbose=False,
                 input_dim=None, output_dim=None, dtype=None):
        """
        :Keyword arguments:
           k
              number of nearest neighbors to use; the node will raise
              an MDPWarning if k is smaller than
              k >= 1 + output_dim + output_dim*(output_dim+1)/2,
              because in this case a less efficient computation must be
              used, and the ablgorithm can become unstable
           r
              regularization constant; as opposed to LLENode, it is
              not possible to compute this constant automatically; it is
              only used during execution
           svd
              if true, use SVD to compute the projection matrix;
              SVD is slower but more stable
           verbose
              if true, displays information about the progress
              of the algorithm
           output_dim
              number of dimensions to output or a float between 0.0
              and 1.0. In the latter case, output_dim specifies the
              desired fraction of variance to be exaplained, and the
              final number of output dimensions is known at the end of
              training (e.g., for 'output_dim=0.95' the algorithm will
              keep as many dimensions as necessary in order to explain
              95% of the input variance)
        """
        LLENode.__init__(self, k, r, svd, verbose,
                         input_dim, output_dim, dtype)

    def _stop_training(self):
        Cumulator._stop_training(self)

        k = self.k
        M = self.data
        N = M.shape[0]

        if k > N:
            err = ('k=%i must be less than'
                   ' or equal to number of training points N=%i' % (k, N))
            raise TrainingException(err)

        if self.verbose:
            print('performing HLLE on %i points in %i dimensions...' % M.shape)

        # determines number of output dimensions: if desired_variance
        # is specified, we need to learn it from the data. Otherwise,
        # it's easy
        learn_outdim = False
        if self.output_dim is None:
            if self.desired_variance is None:
                self.output_dim = self.input_dim
            else:
                learn_outdim = True

        # determine number of output dims, precalculate useful stuff
        if learn_outdim:
            Qs, sig2s, nbrss = self._adjust_output_dim()

        d_out = self.output_dim

        #dp = d_out + (d_out-1) + (d_out-2) + ...
        dp = d_out*(d_out+1)//2

        if min(k, N) <= d_out:
            err = ('k=%i and n=%i (number of input data points) must be'
                   ' larger than output_dim=%i' % (k, N, d_out))
            raise TrainingException(err)

        if k < 1+d_out+dp:
            wrn = ('The number of neighbours, k=%i, is smaller than'
                   ' 1 + output_dim + output_dim*(output_dim+1)/2 = %i,'
                   ' which might result in unstable results.'
                   % (k, 1+d_out+dp))
            _warnings.warn(wrn, MDPWarning)

        #build the weight matrix
        #XXX   for faster implementation, W should be a sparse matrix
        W = numx.zeros((N, dp*N), dtype=self.dtype)

        if self.verbose:
            print(' - constructing [%i x %i] weight matrix...' % W.shape)

        for row in range(N):
            if learn_outdim:
                nbrs = nbrss[row, :]
            else:
                # -----------------------------------------------
                #  find k nearest neighbors
                # -----------------------------------------------
                M_Mi = M-M[row]
                nbrs = numx.argsort((M_Mi**2).sum(1))[1:k+1]

            #-----------------------------------------------
            #  center the neighborhood using the mean
            #-----------------------------------------------
            nbrhd = M[nbrs] # this makes a copy
            nbrhd -= nbrhd.mean(0)

            #-----------------------------------------------
            #  compute local coordinates
            #   using a singular value decomposition
            #-----------------------------------------------
            U, sig, VT = svd(nbrhd)
            nbrhd = U.T[:d_out]
            del VT

            #-----------------------------------------------
            #  build Hessian estimator
            #-----------------------------------------------
            Yi = numx.zeros((dp, k), dtype=self.dtype)
            ct = 0
            for i in range(d_out):
                Yi[ct:ct+d_out-i, :] = nbrhd[i] * nbrhd[i:, :]
                ct += d_out-i
            Yi = numx.concatenate([numx.ones((1, k), dtype=self.dtype),
                                   nbrhd, Yi], 0)

            #-----------------------------------------------
            #  orthogonalize linear and quadratic forms
            #   with QR factorization
            #  and make the weights sum to 1
            #-----------------------------------------------
            if k >= 1+d_out+dp:
                Q, R = numx_linalg.qr(Yi.T)
                w = Q[:, d_out+1:d_out+1+dp]
            else:
                q, r = _mgs(Yi.T)
                w = q[:, -dp:]

            S = w.sum(0) #sum along columns
            #if S[i] is too small, set it equal to 1.0
            # this prevents weights from blowing up
            S[numx.where(numx.absolute(S)<1E-4)] = 1.0
            #print w.shape, S.shape, (w/S).shape
            #print W[nbrs, row*dp:(row+1)*dp].shape
            W[nbrs, row*dp:(row+1)*dp] = old_div(w, S)

        #-----------------------------------------------
        # To find the null space, we want the
        #  first d+1 eigenvectors of W.T*W
        # Compute this using an svd of W
        #-----------------------------------------------

        if self.verbose:
            msg = (' - finding [%i x %i] '
                   'null space of weight matrix...' % (d_out, N))
            print(msg)

        #XXX future work:
        #XXX  use of upcoming ARPACK interface for bottom few eigenvectors
        #XXX   of a sparse matrix will significantly increase the speed
        #XXX   of the next step

        if self.svd:
            sig, U = nongeneral_svd(W.T, range=(2, d_out+1))
            Y = U*numx.sqrt(N)
        else:
            WW = mult(W, W.T)
            # regularizes the eigenvalues, does not change the eigenvectors:
            W_diag_idx = numx.arange(N)
            WW[W_diag_idx, W_diag_idx] += 0.01
            sig, U = symeig(WW, range=(2, self.output_dim+1), overwrite=True)
            Y = U*numx.sqrt(N)
            del WW
        del W

        #-----------------------------------------------
        # Normalize Y
        #
        # Alternative way to do it:
        #  we need R = (Y.T*Y)^(-1/2)
        #   do this with an SVD of Y            del VT

        #      Y = U*sig*V.T
        #      Y.T*Y = (V*sig.T*U.T) * (U*sig*V.T)
        #            = V * (sig*sig.T) * V.T
        #            = V * sig^2 V.T
        #   so
        #      R = V * sig^-1 * V.T
        # The code is:
        #    U, sig, VT = svd(Y)
        #    del U
        #    S = numx.diag(sig**-1)
        #    self.training_projection = mult(Y, mult(VT.T, mult(S, VT)))
        #-----------------------------------------------
        if self.verbose:
            print(' - normalizing null space...')

        C = sqrtm(mult(Y.T, Y))
        self.training_projection = mult(Y, C)



"""
This module contains the basic classes for task processing via a scheduler.
"""
from __future__ import print_function
from builtins import zip
from builtins import object

import threading
import time
import os
try:
    import multiprocessing
except ImportError:
    # Python version < 2.6, have to use fallbacks
    pass


class ResultContainer(object):
    """Abstract base class for result containers."""

    def add_result(self, result_data, task_index):
        """Store a result in the container."""
        pass

    def get_results(self):
        """Return results and reset container."""
        pass


class ListResultContainer(ResultContainer):
    """Basic result container using simply a list."""

    def __init__(self):
        super(ListResultContainer, self).__init__()
        self._results = []

    def add_result(self, result, task_index):
        """Store a result in the container."""
        self._results.append(result)

    def get_results(self):
        """Return the list of results and reset this container.

        Note that the results are stored in the order that they come in, which
        can be different from the orginal task order.
        """
        results = self._results
        self._results = []
        return results


class OrderedResultContainer(ListResultContainer):
    """Default result container with automatic restoring of the result order.

    In general the order of the incoming results in the scheduler can be
    different from the order of the tasks, since some tasks may finish quicker
    than other tasks. This result container restores the original order.
    """

    def __init__(self):
        super(OrderedResultContainer, self).__init__()

    def add_result(self, result, task_index):
        """Store a result in the container.

        The task index is also stored and later used to reconstruct the
        original task order.
        """
        self._results.append((result, task_index))

    def get_results(self):
        """Sort the results into the original order and return them in list."""
        results = self._results
        self._results = []
        results.sort(key=lambda x: x[1])
        return list(zip(*results))[0]


class TaskCallable(object):
    """Abstract base class for task callables.

    This class encapsulates the task behavior and the related fixed data
    (data which stays constant over multiple tasks).
    """

    def setup_environment(self):
        """This hook method is only called when the callable is first called
        in a different Python process / environment.

        It can be used for modifications in the Python environment that are
        required by this callable.
        """
        pass

    def __call__(self, data):
        """Perform the computation and return the result.

        Override this method with a concrete implementation.
        """
        return data

    # TODO: is 'fork' really a good name?
    #    As an alternative one could have a separate CallableFactory class,
    #    but this would make things more complicated for simple callables
    #    (similar to why iterators implement the iterable interface).
    def fork(self):
        """Return a fork of this callable, e.g. by making a copy.

        This method is always called exactly once before a callable is called,
        so instead of the original callable a fresh fork is called. This
        ensures that the original callable is preserved when caching is used.
        If the callable is not modified by the call then it can simply return
        itself.
        """
        return self


class SqrTestCallable(TaskCallable):
    """Callable for testing."""

    def __call__(self, data):
        """Return the squared data."""
        return data**2


class SleepSqrTestCallable(TaskCallable):
    """Callable for testing."""

    def __call__(self, data):
        """Return the squared data[0] after sleeping for data[1] seconds."""
        time.sleep(data[1])
        return data[0]**2


class MDPVersionCallable(TaskCallable):
    """Callable For testing MDP version.

    Should return a unique comparable object which includes version information
    and installed/used modules.
    """

    def __call__(self, data):
        """Ignore input data and return mdp.info()"""
        import mdp
        return mdp.config.info()


class TaskCallableWrapper(TaskCallable):
    """Wrapper to provide a fork method for simple callables like a function.

    This wrapper is applied internally in Scheduler.
    """

    def __init__(self, task_callable):
        """Store and wrap the callable."""
        self._callable = task_callable

    def __call__(self, data):
        """Call the internal callable with the data and return the result."""
        return self._callable(data)


# helper function
def cpu_count():
    """Return the number of CPU cores."""
    try:
        return multiprocessing.cpu_count()
    # TODO: remove except clause once we support only python >= 2.6
    except NameError:
        ## This code part is taken from parallel python.
        # Linux, Unix and MacOS
        if hasattr(os, "sysconf"):
            if "SC_NPROCESSORS_ONLN" in os.sysconf_names:
                # Linux & Unix
                n_cpus = os.sysconf("SC_NPROCESSORS_ONLN")
                if isinstance(n_cpus, int) and n_cpus > 0:
                    return n_cpus
            else:
                # OSX
                return int(os.popen2("sysctl -n hw.ncpu")[1].read())
        # Windows
        if "NUMBER_OF_PROCESSORS" in os.environ:
            n_cpus = int(os.environ["NUMBER_OF_PROCESSORS"])
            if n_cpus > 0:
                return n_cpus
        # Default
        return 1


class Scheduler(object):
    """Base class and trivial implementation for schedulers.

    New tasks are added with add_task(data, callable).
    get_results then returns the results (and locks if tasks are
    pending).

    In this simple scheduler implementation the tasks are simply executed in the
    add_task method.
    """

    def __init__(self, result_container=None, verbose=False):
        """Initialize the scheduler.

        result_container -- Instance of ResultContainer that is used to store
            the results (default is None, in which case a ListResultContainer
            is used).
        verbose -- If True then status messages will be printed to sys.stdout.
        """
        if result_container is None:
            result_container = OrderedResultContainer()
        self.result_container = result_container
        self.verbose = verbose
        self._n_open_tasks = 0  # number of tasks that are currently running
        # count the number of submitted tasks, also used for the task index
        self._task_counter = 0
        self._lock = threading.Lock()
        self._last_callable = None  # last callable is stored
        # task index of the _last_callable, can be *.5 if updated between tasks
        self._last_callable_index = -1.0

    ## public read only properties ##

    @property
    def task_counter(self):
        """This property counts the number of submitted tasks."""
        return self._task_counter

    @property
    def n_open_tasks(self):
        """This property counts of submitted but unfinished tasks."""
        return self._n_open_tasks

    ## main methods ##

    def add_task(self, data, task_callable=None):
        """Add a task to be executed.

        data -- Data for the task.
        task_callable -- A callable, which is called with the data. If it is
            None (default value) then the last provided callable is used.
            If task_callable is not an instance of TaskCallable then a
            TaskCallableWrapper is used.

        The callable together with the data constitutes the task. This method
        blocks if there are no free recources to store or process the task
        (e.g. if no free worker processes are available).
        """
        self._lock.acquire()
        if task_callable is None:
            if self._last_callable is None:
                raise Exception("No task_callable specified and " +
                                "no previous callable available.")
        self._n_open_tasks += 1
        self._task_counter += 1
        task_index = self.task_counter
        if task_callable is None:
            # use the _last_callable_index in _process_task to
            # decide if a cached callable can be used
            task_callable = self._last_callable
        else:
            if not hasattr(task_callable, "fork"):
                # not a TaskCallable (probably a function), so wrap it
                task_callable = TaskCallableWrapper(task_callable)
            self._last_callable = task_callable
            self._last_callable_index = self.task_counter
        self._process_task(data, task_callable, task_index)

    def set_task_callable(self, task_callable):
        """Set the callable that will be used if no task_callable is given.

        Normally the callables are provided via add_task, in which case there
        is no need for this method.

        task_callable -- Callable that will be used unless a new task_callable
            is given.
        """
        self._lock.acquire()
        self._last_callable = task_callable
        # set _last_callable_index to half value since the callable is newer
        # than the last task, but not newer than the next incoming task
        self._last_callable_index = self.task_counter + 0.5
        self._lock.release()

    def _store_result(self, result, task_index):
        """Store a result in the internal result container.

        result -- Result data
        task_index -- Task index. Can be None if an error occured.

        This function blocks to avoid any problems during result storage.
        """
        self._lock.acquire()
        self.result_container.add_result(result, task_index)
        if self.verbose:
            if task_index is not None:
                print("    finished task no. %d" % task_index)
            else:
                print("    task failed")
        self._n_open_tasks -= 1
        self._lock.release()

    def get_results(self):
        """Get the accumulated results from the result container.

        This method blocks if there are open tasks.
        """
        while True:
            self._lock.acquire()
            if self._n_open_tasks == 0:
                results = self.result_container.get_results()
                self._lock.release()
                return results
            else:
                self._lock.release()
                time.sleep(1)

    def shutdown(self):
        """Controlled shutdown of the scheduler.

        This method should always be called when the scheduler is no longer
        needed and before the program shuts down! Otherwise one might get
        error messages.
        """
        self._shutdown()

    ## Context Manager interface ##

    def __enter__(self):
        """Return self."""
        return self

    def __exit__(self, type, value, traceback):
        """Shutdown the scheduler.

        It is important that all the calculations have finished
        when this is called, otherwise the shutdown might fail.
        """
        self.shutdown()

    ## override these methods in custom schedulers ##

    def _process_task(self, data, task_callable, task_index):
        """Process the task and store the result.

        You can override this method for custom schedulers.

        Warning: When this method is entered is has the lock, the lock must be
        released here.

        Warning: Note that fork has not been called yet, so the provided
        task_callable must not be called. Only a forked version can be called.
        """
        # IMPORTANT: always call fork, since it must be called at least once!
        task_callable = task_callable.fork()
        result = task_callable(data)
        # release lock before store_result
        self._lock.release()
        self._store_result(result, task_index)

    def _shutdown(self):
        """Hook method for shutdown to be used in custom schedulers."""
        pass



from builtins import range
from builtins import object
import py.test
import inspect

from mdp import (config, nodes, ClassifierNode,
                 PreserveDimNode, InconsistentDimException)
from ._tools import *

uniform = numx_rand.random

def _rand_labels(x):
    return numx_rand.randint(0, 2, size=(x.shape[0],))

def _rand_labels_array(x):
    return numx_rand.randint(0, 2, size=(x.shape[0], 1))

def _rand_classification_labels_array(x):
    labels = numx_rand.randint(0, 2, size=(x.shape[0],))
    labels[labels==0] = -1
    return labels

def _dumb_quadratic_expansion(x):
    dim_x = x.shape[1]
    return numx.asarray([(x[i].reshape(dim_x,1) *
                          x[i].reshape(1,dim_x)).flatten()
                         for i in range(len(x))])

def _rand_array_halfdim(x):
    return uniform(size=(x.shape[0], x.shape[1]//2))

class Iter(object):
    pass

def _rand_array_single_rows():
    x = uniform((500,4))
    class _Iter(Iter):
        def __iter__(self):
            for row in range(x.shape[0]):
                yield x[numx.newaxis,row,:]
    return _Iter()

def _contrib_get_random_mix():
    return get_random_mix(type='d', mat_dim=(100, 3))[2]

def _positive_get_random_mix():
    return abs(get_random_mix()[2])

def _train_if_necessary(inp, node, sup_arg_gen):
    if node.is_trainable():
        while True:
            if sup_arg_gen is not None:
                # for nodes that need supervision
                node.train(inp, sup_arg_gen(inp))
            else:
                # support generators
                if isinstance(inp, Iter):
                    for x in inp:
                        node.train(x)
                else:
                    node.train(inp)
            if node.get_remaining_train_phase() > 1:
                node.stop_training()
            else:
                break

def _stop_training_or_execute(node, inp):
    if node.is_trainable():
        node.stop_training()
    else:
        if isinstance(inp, Iter):
            for x in inp:
                node.execute(x)
        else:
            node.execute(inp)

def pytest_generate_tests(metafunc):
    generic_test_factory(NODES, metafunc)

def generic_test_factory(big_nodes, metafunc):
    """Generator creating a test for each of the nodes
    based upon arguments in a list of nodes in big_nodes.

    Format of big_nodes:
    each item in the list can be either a
    - class name, in this case the class instances are initialized
      without arguments and default arguments are used during
      the training and execution phases.
    - dict containing items which can override the initialization
      arguments, provide extra arguments for training and/or
      execution.

    Available keys in the configuration dict:
    `klass`
      Mandatory.
      The type of Node.

    `init_args=()`
      A sequence used to provide the initialization arguments to node
      constructor. Before being used, the items in this sequence are
      executed if they are callable. This allows one to create fresh
      instances of nodes before each Node initalization.

    `inp_arg_gen=...a call to get_random_mix('d')`
      Used to construct the `inp` data argument used for training and
      execution. It can be an iterable.

    `sup_arg_gen=None`
      A function taking a single argument (`inp`)
      Used to contruct extra arguments passed to `train`.

    `execute_arg_gen=None`
      A function similar to `sup_arg_gen` but used during execution.
      The return value is unpacked and used as additional arguments to
      `execute`.
    """
    for nodetype in big_nodes:
        if not isinstance(nodetype, dict):
            nodetype = dict(klass=nodetype)
        funcargs = dict(
            init_args=(),
            inp_arg_gen=lambda: get_random_mix(type='d')[2],
            sup_arg_gen=None,
            execute_arg_gen=None)
        funcargs.update(nodetype)

        if hasattr(metafunc.function, 'only_if_node_condition'):
            # A TypeError can be thrown by the condition checking
            # function (e.g. when nodetype.is_trainable() is not a staticmethod).
            condition = metafunc.function.only_if_node_condition
            try:
                if not condition(nodetype['klass']):
                    continue
            except TypeError:
                continue

        theid = nodetype['klass'].__name__
        metafunc.addcall(funcargs, id=theid)

def only_if_node(condition):
    """Execute the test only if condition(nodetype) is True.

    If condition(nodetype) throws TypeError, just assume False.
    """
    def f(func):
        func.only_if_node_condition = condition
        return func
    return f

def call_init_args(init_args):
    return [item() if hasattr(item, '__call__') else item
            for item in init_args]

def test_dtype_consistency(klass, init_args, inp_arg_gen,
                           sup_arg_gen, execute_arg_gen):
    args = call_init_args(init_args)
    supported_types = klass(*args).get_supported_dtypes()
    for dtype in supported_types:
        inp = inp_arg_gen()
        args = call_init_args(init_args)
        node = klass(dtype=dtype, *args)
        _train_if_necessary(inp, node, sup_arg_gen)

        extra = [execute_arg_gen(inp)] if execute_arg_gen else []
        # support generators
        if isinstance(inp, Iter):
            for x in inp:
                out = node.execute(x, *extra)
        else:
            out = node.execute(inp, *extra)
        assert out.dtype == dtype


def test_outputdim_consistency(klass, init_args, inp_arg_gen,
                               sup_arg_gen, execute_arg_gen):
    args = call_init_args(init_args)
    inp = inp_arg_gen()
    # support generators
    if isinstance(inp, Iter):
        for x in inp:
            pass
        output_dim = x.shape[1] // 2
    else:
        output_dim = inp.shape[1] // 2
    extra = [execute_arg_gen(inp)] if execute_arg_gen else []

    def _test(node):
        _train_if_necessary(inp, node, sup_arg_gen)
        # support generators
        if isinstance(inp, Iter):
            for x in inp:
                out = node.execute(x)
        else:
            out = node.execute(inp, *extra)
        assert out.shape[1] == output_dim
        assert node._output_dim == output_dim

    # check if the node output dimension can be set or must be determined
    # by the node
    if (not issubclass(klass, PreserveDimNode) and
        'output_dim' in inspect.getargspec(klass.__init__)[0]):
        # case 1: output dim set in the constructor
        node = klass(output_dim=output_dim, *args)
        _test(node)

        # case 2: output_dim set explicitly
        node = klass(*args)
        node.output_dim = output_dim
        _test(node)
    else:
        if issubclass(klass, PreserveDimNode):
            # check that constructor allows to set output_dim
            assert 'output_dim' in inspect.getargspec(klass.__init__)[0]
            # check that setting the input dim, then incompatible output dims
            # raises an appropriate error
            # case 1: both in the constructor
            py.test.raises(InconsistentDimException,
                   'klass(input_dim=inp.shape[1], output_dim=output_dim, *args)')
            # case 2: first input_dim, then output_dim
            node = klass(input_dim=inp.shape[1], *args)
            py.test.raises(InconsistentDimException,
                           'node.output_dim = output_dim')
            # case 3: first output_dim, then input_dim
            node = klass(output_dim=output_dim, *args)
            node.output_dim = output_dim
            py.test.raises(InconsistentDimException,
                           'node.input_dim = inp.shape[1]')

        # check that output_dim is set to whatever the output dim is
        node = klass(*args)
        _train_if_necessary(inp, node, sup_arg_gen)
        # support generators
        if isinstance(inp, Iter):
            for x in inp:
                out = node.execute(x, *extra)
        else:
            out = node.execute(inp, *extra)

        assert out.shape[1] == node.output_dim

def test_dimdtypeset(klass, init_args, inp_arg_gen,
                     sup_arg_gen, execute_arg_gen):
    init_args = call_init_args(init_args)
    inp = inp_arg_gen()
    node = klass(*init_args)
    _train_if_necessary(inp, node, sup_arg_gen)
    _stop_training_or_execute(node, inp)
    assert node.output_dim is not None
    assert node.dtype is not None
    assert node.input_dim is not None

@only_if_node(lambda nodetype: nodetype.is_invertible())
def test_inverse(klass, init_args, inp_arg_gen,
                 sup_arg_gen, execute_arg_gen):
    args = call_init_args(init_args)
    inp = inp_arg_gen()
    # take the first available dtype for the test
    dtype = klass(*args).get_supported_dtypes()[0]
    args = call_init_args(init_args)
    node = klass(dtype=dtype, *args)
    _train_if_necessary(inp, node, sup_arg_gen)
    extra = [execute_arg_gen(inp)] if execute_arg_gen else []
    out = node.execute(inp, *extra)
    # compute the inverse
    rec = node.inverse(out)
    # cast inp for comparison!
    inp = inp.astype(dtype)
    assert_array_almost_equal_diff(rec, inp, decimal-3)
    assert rec.dtype == dtype

def SFA2Node_inp_arg_gen():
    freqs = [2*numx.pi*100.,2*numx.pi*200.]
    t =  numx.linspace(0, 1, num=1000)
    mat = numx.array([numx.sin(freqs[0]*t),
                      numx.sin(freqs[1]*t)]).T
    inp = mat.astype('d')
    return inp

def NeuralGasNode_inp_arg_gen():
    return numx.asarray([[2.,0,0],[-2,0,0],[0,0,0]])

def LinearRegressionNode_inp_arg_gen():
    return uniform(size=(1000, 5))

def _rand_1d(x):
    return uniform(size=(x.shape[0],))


NODES = [
    dict(klass='NeuralGasNode',
         init_args=[3,NeuralGasNode_inp_arg_gen()],
         inp_arg_gen=NeuralGasNode_inp_arg_gen),
    dict(klass='SFA2Node',
         inp_arg_gen=SFA2Node_inp_arg_gen),
    dict(klass='PolynomialExpansionNode',
         init_args=[3]),
    dict(klass='RBFExpansionNode',
         init_args=[[[0.]*5, [0.]*5], [1., 1.]]),
    dict(klass='GeneralExpansionNode',
         init_args=[[lambda x:x, lambda x: x**2, _dumb_quadratic_expansion]]),
    dict(klass='HitParadeNode',
         init_args=[2, 5]),
    dict(klass='TimeFramesNode',
         init_args=[3, 4]),
    dict(klass='TimeDelayNode',
         init_args=[3, 4]),
    dict(klass='TimeDelaySlidingWindowNode',
         init_args=[3, 4],
         inp_arg_gen=_rand_array_single_rows),
    dict(klass='FDANode',
         sup_arg_gen=_rand_labels),
    dict(klass='GaussianClassifier',
         sup_arg_gen=_rand_labels),
    dict(klass='NearestMeanClassifier',
         sup_arg_gen=_rand_labels),
    dict(klass='KNNClassifier',
         sup_arg_gen=_rand_labels),
    dict(klass='RBMNode',
         init_args=[5]),
    dict(klass='RBMWithLabelsNode',
         init_args=[5, 1],
         sup_arg_gen=_rand_labels_array,
         execute_arg_gen=_rand_labels_array),
    dict(klass='LinearRegressionNode',
         sup_arg_gen=_rand_array_halfdim),
    dict(klass='Convolution2DNode',
         init_args=[mdp.numx.array([[[1.]]]), (5,1)]),
    dict(klass='JADENode',
         inp_arg_gen=_contrib_get_random_mix),
    dict(klass='NIPALSNode',
         inp_arg_gen=_contrib_get_random_mix),
    dict(klass='XSFANode',
         inp_arg_gen=_contrib_get_random_mix,
         init_args=[(nodes.PolynomialExpansionNode, (1,), {}),
                    (nodes.PolynomialExpansionNode, (1,), {}),
                    True]),
    dict(klass='LLENode',
         inp_arg_gen=_contrib_get_random_mix,
         init_args=[3, 0.001, True]),
    dict(klass='HLLENode',
         inp_arg_gen=_contrib_get_random_mix,
         init_args=[10, 0.001, True]),
    dict(klass='KMeansClassifier',
         init_args=[2, 3]),

    dict(klass='PerceptronClassifier',
         sup_arg_gen=_rand_classification_labels_array),
    dict(klass='SimpleMarkovClassifier',
         sup_arg_gen=_rand_classification_labels_array),
    dict(klass='ShogunSVMClassifier',
        sup_arg_gen=_rand_labels_array,
        init_args=["libsvmmulticlass", (), None, "GaussianKernel"]),
    dict(klass='LibSVMClassifier',
        sup_arg_gen=_rand_labels_array,
        init_args=["LINEAR","C_SVC"]),
    dict(klass='MultinomialNBScikitsLearnNode',
         inp_arg_gen=_positive_get_random_mix,
         sup_arg_gen=_rand_labels),
    dict(klass='NeighborsScikitsLearnNode',
        sup_arg_gen=_rand_1d),
    ]

# LabelSpreadingScikitsLearnNode is broken in sklearn version 0.11
# It works fine in version 0.12
EXCLUDE_NODES = ['ICANode', 'LabelSpreadingScikitsLearnNode', 
        'OutputCodeClassifierScikitsLearnNode', 'OneVsOneClassifierScikitsLearnNode',
        'OneVsRestClassifierScikitsLearnNode', 'VotingClassifierScikitsLearnNode']


def generate_nodes_list(nodes_dicts):
    nodes_list = []
    # append nodes with additional arguments or supervised if they exist
    visited = []
    excluded = []
    for dct in nodes_dicts:
        klass = dct['klass']
        if type(klass) is __builtins__['str']:
            # some of the nodes on the list may be optional
            if not hasattr(nodes, klass): continue
            # transform class name into class (needed by automatic tests)
            klass = getattr(nodes, klass)
            dct['klass'] = klass
        # only append to list if the node is present in MDP
        # in case some of the nodes in NODES are optional
        if hasattr(nodes, klass.__name__):
            nodes_list.append(dct)
            visited.append(klass)
    for node_name in EXCLUDE_NODES:
        if hasattr(nodes, node_name):
            excluded.append(getattr(nodes, node_name))
    # append sklearn nodes if supported
    # XXX
    # remove all non classifier nodes from the scikits nodes
    # they do not have a common API that would allow
    # automatic testing
    # XXX
    for node_name in mdp.nodes.__dict__:
        node = mdp.nodes.__dict__[node_name]
        if (inspect.isclass(node)
            and node_name.endswith('ScikitsLearnNode')
            and (node not in visited)
            and (node not in excluded)):
            if issubclass(node, ClassifierNode):
                nodes_list.append(dict(klass=node,
                                       sup_arg_gen=_rand_labels))
                visited.append(node)
            else:
                excluded.append(node)

    # append all other nodes in mdp.nodes
    for attr in dir(nodes):
        if attr[0] == '_':
            continue
        attr = getattr(nodes, attr)
        if (inspect.isclass(attr)
            and issubclass(attr, mdp.Node)
            and attr not in visited
            and attr not in excluded):
            nodes_list.append(attr)
    return nodes_list

NODES = generate_nodes_list(NODES)



import sys
from ._tools import *

def _list_module(module):
    try:
        names = module.__all__
    except AttributeError:
        names = dir(module)
    for name in names:
        if name.startswith('_'):
            continue
        item = getattr(module, name)
        try:
            modname = getattr(item, '__module__')
        except AttributeError:
            continue
        if hasattr(item, '__module__'):
            yield modname, name, item

MODULES = ['mdp',
           'mdp.nodes',
           'mdp.hinet',
           'mdp.parallel',
           'mdp.graph',
           'mdp.utils',
           ]

def pytest_generate_tests(metafunc):
    generate_calls(MODULES, metafunc)

def generate_calls(modules, metafunc):
    for module in modules:
        metafunc.addcall(funcargs=dict(parentname=module), id=module)

def test_exports(parentname):
    rootname = parentname.split('.')[-1]
    module = sys.modules[parentname]
    for modname, itemname, item in _list_module(module):
        parts = modname.split('.')
        assert (parts[0] != rootname or
                modname == parentname), \
                '%s.%s.__module_ == %s != %s' % (
                    parentname, itemname, item.__module__, parentname)



"""
Parallel version of bihinet.
"""
from builtins import range

import mdp
from bimdp.hinet import CloneBiLayer


class ParallelCloneBiLayer(CloneBiLayer, mdp.parallel.ParallelExtensionNode):
    """Parallel version of CloneBiLayer.

    This class also adds support for calling switch_to_instance during training,
    using the join method of the internal nodes.
    """

    def _set_use_copies(self, use_copies):
        """Switch internally between using a single node instance or copies.

        In a normal CloneLayer a single node instance is used to represent all
        the horizontally aligned nodes. But in a BiMDP where the nodes store
        temporary data this may not work.
        Via this method one can therefore create copies of the single node
        instance.

        This method can also be triggered by the use_copies msg key.
        """
        if use_copies and not self.use_copies:
            # switch to node copies
            self.nodes = [self.node.copy() for _ in range(len(self.nodes))]
            self.node = None  # disable single node while copies are used
            self._uses_copies = True
        elif not use_copies and self.use_copies:
            # switch to a single node instance
            if self.is_training():
                for forked_node in self.nodes[1:]:
                    self.nodes[0].join(forked_node)
            elif self.is_bi_training():
                for forked_node in self.nodes[1:]:
                    self.nodes[0].bi_join(forked_node)
            self.node = self.nodes[0]
            self.nodes = (self.node,) * len(self.nodes)

    def _fork(self):
        """Fork the nodes in the layer to fork the layer."""
        forked_node = ParallelCloneBiLayer(
                                node=self.nodes[0].fork(),
                                n_nodes=len(self.nodes),
                                use_copies=False,
                                node_id=self._node_id,
                                dtype=self.get_dtype())
        if self.use_copies:
            # simulate switch_to_copies
            forked_node.nodes = [node.fork() for node in self.nodes]
            forked_node.node = None
            return forked_node
        else:
            return forked_node

    def _join(self, forked_node):
        """Join the trained nodes from the forked layer."""
        if self.use_copies:
            for i_node, layer_node in enumerate(self.nodes):
                layer_node.join(forked_node.nodes[i_node])
        else:
            self.node.join(forked_node.node)
            
    def use_execute_fork(self):
        if self.use_copies:
            return any(node.use_execute_fork() for node in self.nodes)
        else:
            return self.node.use_execute_fork()



from builtins import range
import mdp
from ._tools import *

def test_TimeFramesNode():
    length = 14
    gap = 6
    time_frames = 3
    inp = numx.array([numx.arange(length), -numx.arange(length)]).T
    # create node to be tested
    tf = mdp.nodes.TimeFramesNode(time_frames,gap)
    out = tf.execute(inp)
    # check last element
    assert_equal(out[-1,-1], -length+1)
    # check horizontal sequence
    for i in range(1,time_frames):
        assert_array_equal(out[:,2*i],out[:,0]+i*gap)
        assert_array_equal(out[:,2*i+1],out[:,1]-i*gap)
    # check pseudo-inverse
    rec = tf.pseudo_inverse(out)
    assert_equal(rec.shape[1], inp.shape[1])
    block_size = min(out.shape[0], gap)
    for i in range(0,length,gap):
        assert_array_equal(rec[i:i+block_size], inp[i:i+block_size])

def test_TimeFramesNodeBugInputDim():
    mdp.nodes.TimeFramesNode(time_frames=10, gap=1, input_dim=1)



"""
BiNet version of the htmlvisitor hinet module to convert a flow into HTML.
"""
from builtins import str

import tempfile
import os
import webbrowser

import mdp

from bimdp import BiNode
from bimdp.nodes import SenderBiNode
from bimdp.hinet import CloneBiLayer


class BiHiNetHTMLVisitor(mdp.hinet.HiNetHTMLVisitor):
    """Special version of HiNetHTMLVisitor with BiNode support.

    All bimdp attributes are highligthed via the span.bicolor css tag.
    """

    _BIHINET_STYLE = """
    span.bicolor {
        color: #6633FC;
    }
    """

    @classmethod
    def hinet_css(cls):
        """Return the standard CSS string.

        The CSS should be embedded in the final HTML file.
        """
        css = super(BiHiNetHTMLVisitor, cls).hinet_css()
        return css + cls._BIHINET_STYLE

    def _translate_clonelayer(self, clonelayer):
        """This specialized version checks for CloneBiLayer."""
        f = self._file
        self._open_node_env(clonelayer, "layer")
        f.write('<tr><td class="nodename">')
        f.write(str(clonelayer) + '<br><br>')
        f.write('%d repetitions' % len(clonelayer))
        if isinstance(clonelayer, CloneBiLayer):
            f.write('<br><br>')
            f.write('<span class="bicolor" style="font-size: x-small">')
            f.write('use copies: %s</span>' % str(clonelayer.use_copies))
        f.write('</td>')
        f.write('<td>')
        self._visit_node(clonelayer.nodes[0])
        f.write('</td></tr>')
        self._close_node_env(clonelayer)

    def _write_node_header(self, node, type_id="node"):
        """Write the header content for the node into the HTML file."""
        f = self._file
        if type_id == "flow":
            pass
        elif type_id == "flownode":
            if isinstance(node, BiNode):
                f.write('<tr><td class="dim">')
                f.write('<span class="bicolor">id: %s</span>' % node._node_id)
                f.write('</td></tr>')
        else:
            f.write('<tr><td class="dim">in-dim: %s' % str(node.input_dim))
            if isinstance(node, BiNode):
                f.write('&nbsp;&nbsp;<span class="bicolor">id: %s</span>'
                        % node._node_id)
            f.write('</td></tr>')
        f.write('<tr><td>')
        f.write('<table class="nodestruct">')


@mdp.extension_method("html", SenderBiNode)
def _html_representation(self):
    return ('<span class="bicolor">recipient id: %s </span><br>' %
            str(self._recipient_id))

#@mdp.extension_method("html_representation", BiNode)
#def _html_representation(self):
#    html_repr = super(BiNode, self).html_representation()
#    if self._stop_result:
#        html_repr = [html_repr,
#                     '<span class="bicolor">stop_msg: %s </span><br>'
#                     % str(self._stop_result)]
#    return html_repr

## Helper functions ##

def show_biflow(flow, filename=None, title="MDP flow display",
              show_size=False, browser_open=True):
    """Write a flow with BiMDP nodes into a HTML file, open it in the browser
    and return the file name. Compared the the non-bi function this provides
    special decoration for BiNode attributes.

    flow -- The flow to be shown.
    filename -- Filename for the HTML file to be created. If None
                a temporary file is created.
    title -- Title for the HTML file.
    show_size -- Show the approximate memory footprint of all nodes.
    browser_open -- If True (default value) then the slideshow file is
        automatically opened in a webbrowser.
    """
    if filename is None:
        (fd, filename) = tempfile.mkstemp(suffix=".html", prefix="MDP_")
        html_file = os.fdopen(fd, 'w')
    else:
        html_file = open(filename, 'w')
    html_file.write('<html>\n<head>\n<title>%s</title>\n' % title)
    html_file.write('<style type="text/css" media="screen">')
    html_file.write(mdp.utils.basic_css())
    html_file.write(BiHiNetHTMLVisitor.hinet_css())
    html_file.write('</style>\n</head>\n<body>\n')
    html_file.write('<h3>%s</h3>\n' % title)
    explanation = '(data flows from top to bottom)'
    html_file.write('<par class="explanation">%s</par>\n' % explanation)
    html_file.write('<br><br><br>\n')
    converter = BiHiNetHTMLVisitor(html_file, show_size=show_size)
    converter.convert_flow(flow=flow)
    html_file.write('</body>\n</html>')
    html_file.close()
    if browser_open:
        webbrowser.open(os.path.abspath(filename))
    return filename




from .biflownode import BiFlowNode
from .bilayer import CloneBiLayerException, CloneBiLayer
from .biswitchboard import *
from .bihtmlvisitor import BiHiNetHTMLVisitor, show_biflow

del biflownode
del bilayer
del biswitchboard
del bihtmlvisitor



from builtins import object

import mdp
import sys
import py.test

def teardown_function(function):
    """Deactivate all extensions and remove testing extensions."""
    mdp.deactivate_extensions(mdp.get_active_extensions())
    for key in mdp.get_extensions().copy():
        if key.startswith("__test"):
            del mdp.get_extensions()[key]

def testSimpleExtension():
    """Test for a single new extension."""
    class TestExtensionNode(mdp.ExtensionNode):
        extension_name = "__test"
        def _testtest(self):
            pass
        _testtest_attr = 1337
    class TestSFANode(TestExtensionNode, mdp.nodes.SFANode):
        def _testtest(self):
            return 42
        _testtest_attr = 1338
    sfa_node = mdp.nodes.SFANode()
    mdp.activate_extension("__test")
    assert sfa_node._testtest() == 42
    assert sfa_node._testtest_attr == 1338
    mdp.deactivate_extension("__test")
    assert not hasattr(mdp.nodes.SFANode, "_testtest")

def testContextDecorator():
    """Test the with_extension function decorator."""

    class Test1ExtensionNode(mdp.ExtensionNode):
        extension_name = "__test1"
        def _testtest(self):
            pass

    @mdp.with_extension("__test1")
    def test():
        return mdp.get_active_extensions()

    # check that the extension is activated
    assert mdp.get_active_extensions() == []
    active = test()
    assert active == ["__test1"]
    assert mdp.get_active_extensions() == []

    # check that it is only deactiveted if it was activated there
    mdp.activate_extension("__test1")
    active = test()
    assert active == ["__test1"]
    assert mdp.get_active_extensions() == ["__test1"]

def testContextManager1():
    """Test that the context manager activates extensions."""
    class Test1ExtensionNode(mdp.ExtensionNode):
        extension_name = "__test1"
        def _testtest(self):
            pass
    class Test2ExtensionNode(mdp.ExtensionNode):
        extension_name = "__test2"
        def _testtest(self):
            pass
    assert mdp.get_active_extensions() == []
    with mdp.extension('__test1'):
        assert mdp.get_active_extensions() == ['__test1']
    assert mdp.get_active_extensions() == []
    # with multiple extensions
    with mdp.extension(['__test1', '__test2']):
        active = mdp.get_active_extensions()
        assert '__test1' in active
        assert '__test2' in active
    assert mdp.get_active_extensions() == []
    mdp.activate_extension("__test1")
    # Test that only activated extensions are deactiveted.
    with mdp.extension(['__test1', '__test2']):
        active = mdp.get_active_extensions()
        assert '__test1' in active
        assert '__test2' in active
    assert mdp.get_active_extensions() == ["__test1"]

def testDecoratorExtension():
    """Test extension decorator with a single new extension."""
    @mdp.extension_method("__test", mdp.nodes.SFANode, "_testtest")
    def _sfa_testtest(self):
        return 42
    @mdp.extension_method("__test", mdp.nodes.SFA2Node)
    def _testtest(self):
        return 42 + _sfa_testtest(self)
    sfa_node = mdp.nodes.SFANode()
    sfa2_node = mdp.nodes.SFA2Node()
    mdp.activate_extension("__test")
    assert sfa_node._testtest() == 42
    assert sfa2_node._testtest() == 84
    mdp.deactivate_extension("__test")
    assert not hasattr(mdp.nodes.SFANode, "_testtest")
    assert not hasattr(mdp.nodes.SFA2Node, "_testtest")

def testDecoratorInheritance():
    """Test inhertiance with decorators for a single new extension."""
    class TestExtensionNode(mdp.ExtensionNode):
        extension_name = "__test"
        def _testtest(self):
            pass
    @mdp.extension_method("__test", mdp.nodes.SFANode, "_testtest")
    def _sfa_testtest(self):
        return 42
    @mdp.extension_method("__test", mdp.nodes.SFA2Node)
    def _testtest(self):
        return 42 + super(mdp.nodes.SFA2Node, self)._testtest()
    sfa_node = mdp.nodes.SFANode()
    sfa2_node = mdp.nodes.SFA2Node()
    mdp.activate_extension("__test")
    assert sfa_node._testtest() == 42
    assert sfa2_node._testtest() == 84

def testExtensionInheritance():
    """Test inheritance of extension nodes."""
    class TestExtensionNode(mdp.ExtensionNode):
        extension_name = "__test"
        def _testtest(self):
            pass
    class TestSFANode(TestExtensionNode, mdp.nodes.SFANode):
        def _testtest(self):
            return 42
        _testtest_attr = 1337
    class TestSFA2Node(TestSFANode, mdp.nodes.SFA2Node):
        def _testtest(self):
            if sys.version_info[0] < 3:
                return TestSFANode._testtest.__func__(self)
            else:
                return TestSFANode._testtest(self)
    sfa2_node = mdp.nodes.SFA2Node()
    mdp.activate_extension("__test")
    assert sfa2_node._testtest() == 42
    assert sfa2_node._testtest_attr == 1337

def testExtensionInheritance2():
    """Test inheritance of extension nodes, using super."""
    class TestExtensionNode(mdp.ExtensionNode):
        extension_name = "__test"
        def _testtest(self):
            pass
    class TestSFANode(TestExtensionNode, mdp.nodes.SFANode):
        def _testtest(self):
            return 42
    class TestSFA2Node(mdp.nodes.SFA2Node, TestSFANode):
        def _testtest(self):
            return super(mdp.nodes.SFA2Node, self)._testtest()
    sfa2_node = mdp.nodes.SFA2Node()
    mdp.activate_extension("__test")
    assert sfa2_node._testtest() == 42

def testExtensionInheritance3():
    """Test explicit use of extension nodes and inheritance."""
    class TestExtensionNode(mdp.ExtensionNode):
        extension_name = "__test"
        def _testtest(self):
            pass
    class TestSFANode(TestExtensionNode, mdp.nodes.SFANode):
        def _testtest(self):
            return 42
    # Note the inheritance order, otherwise this would not work.
    class TestSFA2Node(mdp.nodes.SFA2Node, TestSFANode):
        def _testtest(self):
            return super(mdp.nodes.SFA2Node, self)._testtest()
    sfa2_node = TestSFA2Node()
    assert sfa2_node._testtest() == 42

def testMultipleExtensions():
    """Test behavior of multiple extensions."""
    class Test1ExtensionNode(mdp.ExtensionNode, mdp.Node):
        extension_name = "__test1"
        def _testtest1(self):
            pass
    class Test2ExtensionNode(mdp.ExtensionNode, mdp.Node):
        extension_name = "__test2"
        def _testtest2(self):
            pass
    mdp.activate_extension("__test1")
    node = mdp.Node()
    node._testtest1()
    mdp.activate_extension("__test2")
    node._testtest2()
    mdp.deactivate_extension("__test1")
    assert not hasattr(mdp.nodes.SFANode, "_testtest1")
    mdp.activate_extension("__test1")
    node._testtest1()
    mdp.deactivate_extensions(["__test1", "__test2"])
    assert not hasattr(mdp.nodes.SFANode, "_testtest1")
    assert not hasattr(mdp.nodes.SFANode, "_testtest2")

def testExtCollision():
    """Test the check for method name collision."""
    class Test1ExtensionNode(mdp.ExtensionNode, mdp.Node):
        extension_name = "__test1"
        def _testtest(self):
            pass
    class Test2ExtensionNode(mdp.ExtensionNode, mdp.Node):
        extension_name = "__test2"
        def _testtest(self):
            pass
    py.test.raises(mdp.ExtensionException,
                   mdp.activate_extensions, ["__test1", "__test2"])
    # none of the extension should be active after the exception
    assert not hasattr(mdp.Node, "_testtest")

def testExtensionInheritanceInjection():
    """Test the injection of inherited methods"""
    class TestNode(object):
        def _test1(self):
            return 0
    class TestExtensionNode(mdp.ExtensionNode):
        extension_name = "__test"
        def _test1(self):
            return 1
        def _test2(self):
            return 2
        def _test3(self):
            return 3
    class TestNodeExt(TestExtensionNode, TestNode):
        def _test2(self):
            return "2b"
    @mdp.extension_method("__test", TestNode)
    def _test4(self):
        return 4
    test_node = TestNode()
    mdp.activate_extension("__test")
    assert test_node._test1() == 1
    assert test_node._test2() == "2b"
    assert test_node._test3() == 3
    assert test_node._test4() == 4
    mdp.deactivate_extension("__test")
    assert test_node._test1() == 0
    assert not hasattr(test_node, "_test2")
    assert not hasattr(test_node, "_test3")
    assert not hasattr(test_node, "_test4")

def testExtensionInheritanceInjectionNonExtension():
    """Test non_extension method injection."""
    class TestExtensionNode(mdp.ExtensionNode):
        extension_name = "__test"
        def _execute(self):
            return 0
    class TestNode(mdp.Node):
        # no _execute method
        pass
    class ExtendedTestNode(TestExtensionNode, TestNode):
        pass
    test_node = TestNode()
    mdp.activate_extension('__test')
    assert hasattr(test_node, "_non_extension__execute")
    mdp.deactivate_extension('__test')
    assert not hasattr(test_node, "_non_extension__execute")
    assert not hasattr(test_node, "_extension_for__execute")
    # test that the non-native _execute has been completely removed
    assert "_execute" not in test_node.__class__.__dict__

def testExtensionInheritanceInjectionNonExtension2():
    """Test non_extension method injection."""
    class TestExtensionNode(mdp.ExtensionNode):
        extension_name = "__test"
        def _execute(self):
            return 0
    class TestNode(mdp.Node):
        def _execute(self):
            return 1
    class ExtendedTestNode(TestExtensionNode, TestNode):
        pass
    test_node = TestNode()
    mdp.activate_extension('__test')
    # test that non-extended attribute has been added as well
    assert hasattr(test_node, "_non_extension__execute")
    mdp.deactivate_extension('__test')
    assert not hasattr(test_node, "_non_extension__execute")
    assert not hasattr(test_node, "_extension_for__execute")
    # test that the native _execute has been preserved
    assert "_execute" in test_node.__class__.__dict__

def testExtensionInheritanceTwoExtensions():
    """Test non_extension injection for multiple extensions."""
    class Test1ExtensionNode(mdp.ExtensionNode):
        extension_name = "__test1"
        def _execute(self):
            return 1
    class Test2ExtensionNode(mdp.ExtensionNode):
        extension_name = "__test2"
    class Test3ExtensionNode(mdp.ExtensionNode):
        extension_name = "__test3"
        def _execute(self):
            return "3a"
    class TestNode1(mdp.Node):
        pass
    class TestNode2(TestNode1):
        pass
    class ExtendedTest1Node2(Test1ExtensionNode, TestNode2):
        pass
    class ExtendedTest2Node1(Test2ExtensionNode, TestNode1):
        def _execute(self):
            return 2
    class ExtendedTest3Node1(Test3ExtensionNode, TestNode1):
        def _execute(self):
            return "3b"
    test_node = TestNode2()
    mdp.activate_extension('__test2')
    assert test_node._execute() == 2
    mdp.deactivate_extension('__test2')
    # in this order TestNode2 should get execute from __test1,
    # the later addition by __test1 to TestNode1 doesn't matter
    mdp.activate_extensions(['__test1', '__test2'])
    assert test_node._execute() == 1
    mdp.deactivate_extensions(['__test2', '__test1'])
    # now activate in inverse order
    # TestNode2 already gets _execute from __test2, but that is still
    # overriden by __test1, thats how its registered in _extensions
    mdp.activate_extensions(['__test2', '__test1'])
    assert test_node._execute() == 1
    mdp.deactivate_extensions(['__test2', '__test1'])
    ## now the same with extension 3
    mdp.activate_extension('__test3')
    assert test_node._execute() == "3b"
    mdp.deactivate_extension('__test3')
    # __test3 does not override, since the _execute slot for Node2
    # was first filled by __test1
    mdp.activate_extensions(['__test3', '__test1'])
    assert test_node._execute() == 1
    mdp.deactivate_extensions(['__test3', '__test1'])
    # inverse order
    mdp.activate_extensions(['__test1', '__test3'])
    assert test_node._execute() == 1
    mdp.deactivate_extensions(['__test2', '__test1'])
    
def testExtensionSetupTeardown():
    """Test defining setup and teardown functions."""
    setup_calls = []
    teardown_calls = []
    @mdp.extension_setup("__test")
    def dummy_setup():
        setup_calls.append(True)
    @mdp.extension_teardown("__test")
    def dummy_setup():
        teardown_calls.append(True)
    mdp.activate_extension("__test")
    assert len(setup_calls) == 1
    mdp.deactivate_extension("__test")
    assert len(teardown_calls) == 1

def testExtensionDuplicateSetup():
    """Test that you can define the setup function only once."""
    def dummy_setup1():
        pass
    def dummy_setup2():
        pass
    mdp.extension_setup("__test")(dummy_setup1)
    py.test.raises(mdp.ExtensionException,
                   lambda: mdp.extension_setup("__test")(dummy_setup2))

def testExtensionDuplicateTeardown():
    """Test that you can define the teardown function only once."""
    def dummy_setup1():
        pass
    def dummy_setup2():
        pass
    mdp.extension_teardown("__test")(dummy_setup1)
    py.test.raises(mdp.ExtensionException,
                   lambda: mdp.extension_teardown("__test")(dummy_setup2))



from __future__ import division
from builtins import range
from past.utils import old_div
import mdp
from ._tools import *

requires_scipy = skip_on_condition(
    "not mdp.numx_description == 'scipy'",
    "This test requires 'scipy'")

def dumb_quadratic_expansion(x):
    dim_x = x.shape[1]
    return numx.asarray([(x[i].reshape(dim_x,1) *
                          x[i].reshape(1,dim_x)).flatten()
                         for i in range(len(x))])

def testGeneralExpansionNode():
    samples = 2
    input_dim = 10
    funcs = [lambda x:x, lambda x: x**2, dumb_quadratic_expansion]

    cen = mdp.nodes.GeneralExpansionNode(funcs)

    input = numx.random.normal(size=(samples, input_dim))
    out = cen.execute(input)
    assert_array_almost_equal(out[:, 0:input_dim], input,
                              6, "incorrect identity expansion")
    assert_array_almost_equal(out[:, input_dim:2*input_dim],
                              input ** 2, 6, "incorrect squares expansion")
    assert_array_almost_equal(out[:, 2*input_dim:],
                              dumb_quadratic_expansion(input), 6,
                              "incorrect quadratic expansion")
    assert cen.expanded_dim(input_dim) == 2 * input_dim + input_dim**2, "expanded_dim failed"
    assert_array_almost_equal(cen.output_sizes(input_dim),
                              numx.array([input_dim, input_dim,
                                          input_dim*input_dim]), 6,
                              "output_sizes failed")

@requires_scipy
def testGeneralExpansionNode_inverse():
    samples = 2
    input_dim = 10
    funcs = [lambda x:x, lambda x: x**2, dumb_quadratic_expansion]

    cen = mdp.nodes.GeneralExpansionNode(funcs)
    input = numx.random.normal(size=(samples, input_dim))
    out = cen.execute(input)
    app_input = cen.pseudo_inverse(out, use_hint=True)
    assert_array_almost_equal_diff(input, app_input, 6,
                                   'inversion not good enough by use_hint=True')

    # ???
    # testing with use_hint = False is tricky, it often fails.
    # we try 20 times in a row and hope for the best
    trials = 20
    for trial in range(trials):
        cen = mdp.nodes.GeneralExpansionNode(funcs)
        input = numx.random.normal(size=(samples, input_dim))
        out = cen.execute(input)
        app_input = cen.pseudo_inverse(out, use_hint=False)
        maxdiff = old_div(max(numx.ravel(abs(app_input-input))),\
                  max(max(abs(numx.ravel(app_input))),max(abs(numx.ravel(input)))))
        cond = maxdiff < 10** (-4)
        if cond:
            break
    assert cond, 'inversion not good enough by use_hint=False'




from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
__docformat__ = "restructuredtext en"

import mdp
from mdp import numx, numx_linalg, utils, NodeException
from mdp.utils import mult, CovarianceMatrix
import warnings

sqrt, inv, det = numx.sqrt, utils.inv, numx_linalg.det
normal = mdp.numx_rand.normal

# decreasing likelihood message
_LHOOD_WARNING = ('Likelihood decreased in FANode. This is probably due '
                  'to some numerical errors.')
warnings.filterwarnings('always', _LHOOD_WARNING, mdp.MDPWarning)

class FANode(mdp.Node):
    """Perform Factor Analysis.

    The current implementation should be most efficient for long
    data sets: the sufficient statistics are collected in the
    training phase, and all EM-cycles are performed at
    its end.

    The ``execute`` method returns the Maximum A Posteriori estimate
    of the latent variables. The ``generate_input`` method generates
    observations from the prior distribution.

    **Internal variables of interest**

      ``self.mu``
          Mean of the input data (available after training)

      ``self.A``
          Generating weights (available after training)

      ``self.E_y_mtx``
          Weights for Maximum A Posteriori inference

      ``self.sigma``
          Vector of estimated variance of the noise
          for all input components

    More information about Factor Analysis can be found in
    Max Welling's classnotes:
    http://www.ics.uci.edu/~welling/classnotes/classnotes.html ,
    in the chapter 'Linear Models'.
    """
    def __init__(self, tol=1e-4, max_cycles=100, verbose=False,
                 input_dim=None, output_dim=None, dtype=None):

        """
        :Parameters:
          tol
            tolerance (minimum change in log-likelihood before exiting
            the EM algorithm)
          max_cycles
            maximum number of EM cycles
          verbose
            if true, print log-likelihood during the EM-cycles
        """
        # Notation as in Max Welling's notes
        super(FANode, self).__init__(input_dim, output_dim, dtype)
        self.tol = tol
        self.max_cycles = max_cycles
        self.verbose = verbose
        self._cov_mtx = CovarianceMatrix(dtype, bias=True)

    def _train(self, x):
        # update the covariance matrix
        self._cov_mtx.update(x)

    def _stop_training(self):
        #### some definitions
        verbose = self.verbose
        typ = self.dtype
        tol = self.tol
        d = self.input_dim
        # if the number of latent variables is not specified,
        # set it equal to the number of input components
        if not self.output_dim:
            self.output_dim = d
        k = self.output_dim
        # indices of the diagonal elements of a dxd or kxk matrix
        idx_diag_d = [i*(d+1) for i in range(d)]
        idx_diag_k = [i*(k+1) for i in range(k)]
        # constant term in front of the log-likelihood
        const = -d/2. * numx.log(2.*numx.pi)

        ##### request the covariance matrix and clean up
        cov_mtx, mu, tlen = self._cov_mtx.fix()
        del self._cov_mtx
        cov_diag = cov_mtx.diagonal()

        ##### initialize the parameters
        # noise variances
        sigma = cov_diag
        # loading factors
        # Zoubin uses the determinant of cov_mtx^1/d as scale but it's
        # too slow for large matrices. Is the product of the diagonal a good
        # approximation?
        if d<=300:
            scale = det(cov_mtx)**(old_div(1.,d))
        else:
            scale = numx.product(sigma)**(old_div(1.,d))
        if scale <= 0.:
            err = ("The covariance matrix of the data is singular. "
                   "Redundant dimensions need to be removed.")
            raise NodeException(err)

        A = normal(0., sqrt(old_div(scale,k)), size=(d, k)).astype(typ)

        ##### EM-cycle
        lhood_curve = []
        base_lhood = None
        old_lhood = -numx.inf
        for t in range(self.max_cycles):
            ## compute B = (A A^T + Sigma)^-1
            B = mult(A, A.T)
            # B += diag(sigma), avoid computing diag(sigma) which is dxd
            B.ravel().put(idx_diag_d, B.ravel().take(idx_diag_d)+sigma)
            # this quantity is used later for the log-likelihood
            # abs is there to avoid numerical errors when det < 0
            log_det_B = numx.log(abs(det(B)))
            # end the computation of B
            B = inv(B)

            ## other useful quantities
            trA_B = mult(A.T, B)
            trA_B_cov_mtx = mult(trA_B, cov_mtx)

            ##### E-step
            ## E_yyT = E(y_n y_n^T | x_n)
            E_yyT = - mult(trA_B, A) + mult(trA_B_cov_mtx, trA_B.T)
            # E_yyT += numx.eye(k)
            E_yyT.ravel().put(idx_diag_k, E_yyT.ravel().take(idx_diag_k)+1.)

            ##### M-step
            A = mult(trA_B_cov_mtx.T, inv(E_yyT))
            sigma = cov_diag - (mult(A, trA_B_cov_mtx)).diagonal()

            ##### log-likelihood
            trace_B_cov = (B*cov_mtx.T).sum()
            # this is actually likelihood/tlen.
            lhood = const - 0.5*log_det_B - 0.5*trace_B_cov
            if verbose:
                print('cycle', t, 'log-lhood:', lhood)

            ##### convergence criterion
            if base_lhood is None:
                base_lhood = lhood
            else:
                # convergence criterion
                if (lhood-base_lhood)<(1.+tol)*(old_lhood-base_lhood):
                    break
                if lhood < old_lhood:
                    # this should never happen
                    # it sometimes does, e.g. if the noise is extremely low,
                    # because of numerical rounding effects
                    warnings.warn(_LHOOD_WARNING, mdp.MDPWarning)
            old_lhood = lhood
            lhood_curve.append(lhood)

        self.tlen = tlen
        self.A = A
        self.mu = mu.reshape(1, d)
        self.sigma = sigma

        ## MAP matrix
        # compute B = (A A^T + Sigma)^-1
        B = mult(A, A.T).copy()
        B.ravel().put(idx_diag_d, B.ravel().take(idx_diag_d)+sigma)
        B = inv(B)
        self.E_y_mtx = mult(B.T, A)

        self.lhood = lhood_curve

    def _execute(self, x):
        return mult(x-self.mu, self.E_y_mtx)

    @staticmethod
    def is_invertible():
        return False

    def generate_input(self, len_or_y=1, noise=False):
        """
        Generate data from the prior distribution.

        If the training phase has not been completed yet, call stop_training.

        :Arguments:
          len_or_y
                    If integer, it specified the number of observation
                    to generate. If array, it is used as a set of samples
                    of the latent variables
          noise
                    if true, generation includes the estimated noise
        """

        self._if_training_stop_training()

        # set the output dimension if necessary
        if self.output_dim is None:
            # if the input_dim is not defined, raise an exception
            if self.input_dim is None:
                errstr = ("Number of input dimensions undefined. Inversion "
                          "not possible.")
                raise NodeException(errstr)
            self.output_dim = self.input_dim

        if isinstance(len_or_y, int):
            size = (len_or_y, self.output_dim)
            y = self._refcast(mdp.numx_rand.normal(size=size))
        else:
            y = self._refcast(len_or_y)
            self._check_output(y)

        res = mult(y, self.A.T)+self.mu
        if noise:
            ns = mdp.numx_rand.normal(size=(y.shape[0], self.input_dim))
            ns *= numx.sqrt(self.sigma)
            res += self._refcast(ns)
        return res



from __future__ import division
from past.utils import old_div
from ._tools import *

requires_scikits = skip_on_condition(
    "not mdp.config.has_sklearn or mdp.numx_description != 'scipy'",
    "This test requires sklearn and SciPy")

requires_pcasikitslearnnode = skip_on_condition(
    "'PCAScikitsLearnNode' not in dir(mdp.nodes)",
    "This test requires sklearn.decomposition.pca.PCA to be available")

@requires_scikits
@requires_pcasikitslearnnode
def test_scikits_PCANode_training():
    """Check functionality of scikits' PCANode."""
    node = mdp.nodes.PCAScikitsLearnNode(n_components=2)

    # the first two principal components are the second and fourth axes
    T = 50000
    x = numx_rand.randn(T, 4)
    x[:,1] *= 10.
    x[:,3] *= 100.

    node.train(x)
    node.stop_training()
    y = node.execute(x)

    # check dimensionality
    assert y.shape[1] == 2
    assert y.shape[0] == T

    # arrays should be equal up to sign
    if (y[:,0]*x[:,3]).mean() < 0.: y[:,0] *= -1.
    if (y[:,1]*x[:,1]).mean() < 0.: y[:,1] *= -1.
    assert_array_almost_equal(old_div(y[:,0],100.), old_div(x[:,3],100.), 1)
    assert_array_almost_equal(old_div(y[:,1],10.), old_div(x[:,1],10.), 1)




from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from builtins import object
import mdp
import sys as _sys
import os as _os
import inspect as _inspect
import warnings as _warnings
import traceback as _traceback
import pickle as _cPickle
import tempfile as _tempfile
import copy as _copy

from mdp import numx

class CrashRecoveryException(mdp.MDPException):
    """Class to handle crash recovery """
    def __init__(self, *args):
        """Allow crash recovery.
        Arguments: (error_string, crashing_obj, parent_exception)
        The crashing object is kept in self.crashing_obj
        The triggering parent exception is kept in self.parent_exception.
        """
        errstr = args[0]
        self.crashing_obj = args[1]
        self.parent_exception = args[2]
        # ?? python 2.5: super(CrashRecoveryException, self).__init__(errstr)
        mdp.MDPException.__init__(self, errstr)

    def dump(self, filename=None):
        """
        Save a pickle dump of the crashing object on filename.
        If filename is None, the crash dump is saved on a file created by
        the tempfile module.
        Return the filename.
        """
        if filename is None:
            # This 'temporary file' should actually stay 'forever', i.e. until
            # deleted by the user.
            (fd, filename)=_tempfile.mkstemp(suffix=".pic", prefix="MDPcrash_")
            fl = _os.fdopen(fd, 'w+b', -1)
        else:
            fl = open(filename, 'w+b', -1)
        _cPickle.dump(self.crashing_obj, fl)
        fl.close()
        return filename

class FlowException(mdp.MDPException):
    """Base class for exceptions in Flow subclasses."""
    pass

class FlowExceptionCR(CrashRecoveryException, FlowException):
    """Class to handle flow-crash recovery """

    def __init__(self, *args):
        """Allow crash recovery.

        Arguments: (error_string, flow_instance, parent_exception)
        The triggering parent exception is kept in self.parent_exception.
        If flow_instance._crash_recovery is set, save a crash dump of
        flow_instance on the file self.filename"""
        CrashRecoveryException.__init__(self, *args)
        rec = self.crashing_obj._crash_recovery
        errstr = args[0]
        if rec:
            if isinstance(rec, str):
                name = rec
            else:
                name = None
            name = CrashRecoveryException.dump(self, name)
            dumpinfo = '\nA crash dump is available on: "%s"' % name
            self.filename = name
            errstr = errstr+dumpinfo

        Exception.__init__(self, errstr)

class Flow(object):
    """A 'Flow' is a sequence of nodes that are trained and executed
    together to form a more complex algorithm.  Input data is sent to the
    first node and is successively processed by the subsequent nodes along
    the sequence.

    Using a flow as opposed to handling manually a set of nodes has a
    clear advantage: The general flow implementation automatizes the
    training (including supervised training and multiple training phases),
    execution, and inverse execution (if defined) of the whole sequence.

    Crash recovery is optionally available: in case of failure the current
    state of the flow is saved for later inspection. A subclass of the
    basic flow class ('CheckpointFlow') allows user-supplied checkpoint
    functions to be executed at the end of each phase, for example to save
    the internal structures of a node for later analysis.
    Flow objects are Python containers. Most of the builtin 'list'
    methods are available. A 'Flow' can be saved or copied using the
    corresponding 'save' and 'copy' methods.
    """

    def __init__(self, flow, crash_recovery=False, verbose=False):
        """
        Keyword arguments:

        flow -- a list of Nodes
        crash_recovery -- set (or not) Crash Recovery Mode (save node
                          in case a failure)
        verbose -- if True, print some basic progress information
        """
        self._check_nodes_consistency(flow)
        self.flow = flow
        self.verbose = verbose
        self.set_crash_recovery(crash_recovery)

    def _propagate_exception(self, except_, nodenr):
        # capture exception. the traceback of the error is printed and a
        # new exception, containing the identity of the node in the flow
        # is raised. Allow crash recovery.
        (etype, val, tb) = _sys.exc_info()
        prev = ''.join(_traceback.format_exception(except_.__class__,
                                                   except_,tb))
        act = "\n! Exception in node #%d (%s):\n" % (nodenr,
                                                     str(self.flow[nodenr]))
        errstr = ''.join(('\n', 40*'-', act, 'Node Traceback:\n', prev, 40*'-'))
        raise FlowExceptionCR(errstr, self, except_)

    def _train_node(self, data_iterable, nodenr):
        """Train a single node in the flow.

        nodenr -- index of the node in the flow
        """
        node = self.flow[nodenr]
        if (data_iterable is not None) and (not node.is_trainable()):
            # attempted to train a node although it is not trainable.
            # raise a warning and continue with the next node.
            # wrnstr = "\n! Node %d is not trainable" % nodenr + \
            #        "\nYou probably need a 'None' iterable for"+\
            #         " this node. Continuing anyway."
            #_warnings.warn(wrnstr, mdp.MDPWarning)
            return
        elif (data_iterable is None) and node.is_training():
            # None instead of iterable is passed to a training node
            err_str = ("\n! Node %d is training"
                       " but instead of iterable received 'None'." % nodenr)
            raise FlowException(err_str)
        elif (data_iterable is None) and (not node.is_trainable()):
            # skip training if node is not trainable
            return

        try:
            train_arg_keys = self._get_required_train_args(node)
            train_args_needed = bool(len(train_arg_keys))
            ## We leave the last training phase open for the
            ## CheckpointFlow class.
            ## Checkpoint functions must close it explicitly if needed!
            ## Note that the last training_phase is closed
            ## automatically when the node is executed.
            while True:
                empty_iterator = True
                for x in data_iterable:
                    empty_iterator = False
                    # the arguments following the first are passed only to the
                    # currently trained node, allowing the implementation of
                    # supervised nodes
                    if (type(x) is tuple) or (type(x) is list):
                        arg = x[1:]
                        x = x[0]
                    else:
                        arg = ()
                    # check if the required number of arguments was given
                    if train_args_needed:
                        if len(train_arg_keys) != len(arg):
                            err = ("Wrong number of arguments provided by " +
                                   "the iterable for node #%d " % nodenr +
                                   "(%d needed, %d given).\n" %
                                   (len(train_arg_keys), len(arg)) +
                                   "List of required argument keys: " +
                                   str(train_arg_keys))
                            raise FlowException(err)
                    # filter x through the previous nodes
                    if nodenr > 0:
                        x = self._execute_seq(x, nodenr-1)
                    # train current node
                    node.train(x, *arg)
                if empty_iterator:
                    if node.get_current_train_phase() == 1:
                        err_str = ("The training data iteration for node "
                                   "no. %d could not be repeated for the "
                                   "second training phase, you probably "
                                   "provided an iterator instead of an "
                                   "iterable." % (nodenr+1))
                        raise FlowException(err_str)
                    else:
                        err_str = ("The training data iterator for node "
                                   "no. %d is empty." % (nodenr+1))
                        raise FlowException(err_str)
                self._stop_training_hook()
                if node.get_remaining_train_phase() > 1:
                    # close the previous training phase
                    node.stop_training()
                else:
                    break
        except mdp.TrainingFinishedException as e:
            # attempted to train a node although its training phase is already
            # finished. raise a warning and continue with the next node.
            wrnstr = ("\n! Node %d training phase already finished"
                      " Continuing anyway." % nodenr)
            _warnings.warn(wrnstr, mdp.MDPWarning)
        except FlowExceptionCR as e:
            # this exception was already propagated,
            # probably during the execution  of a node upstream in the flow
            (exc_type, val) = _sys.exc_info()[:2]
            prev = ''.join(_traceback.format_exception_only(e.__class__, e))
            prev = prev[prev.find('\n')+1:]
            act = "\nWhile training node #%d (%s):\n" % (nodenr,
                                                         str(self.flow[nodenr]))
            err_str = ''.join(('\n', 40*'=', act, prev, 40*'='))
            raise FlowException(err_str)
        except Exception as e:
            # capture any other exception occured during training.
            self._propagate_exception(e, nodenr)

    def _stop_training_hook(self):
        """Hook method that is called before stop_training is called."""
        pass

    @staticmethod
    def _get_required_train_args(node):
        """Return arguments in addition to self and x for node.train.

        Argumentes that have a default value are ignored.
        """
        train_arg_spec = _inspect.getargspec(node._train)
        train_arg_keys = train_arg_spec[0][2:]  # ignore self, x
        if train_arg_spec[3]:
            # subtract arguments with a default value
            train_arg_keys = train_arg_keys[:-len(train_arg_spec[3])]
        return train_arg_keys

    def _train_check_iterables(self, data_iterables):
        """Return the data iterables after some checks and sanitizing.

        Note that this method does not distinguish between iterables and
        iterators, so this must be taken care of later.
        """
        # verifies that the number of iterables matches that of
        # the signal nodes and multiplies them if needed.
        flow = self.flow

        # if a single array is given wrap it in a list of lists,
        # note that a list of 2d arrays is not valid
        if isinstance(data_iterables, numx.ndarray):
            data_iterables = [[data_iterables]] * len(flow)

        if not isinstance(data_iterables, list):
            err_str = ("'data_iterables' must be either a list of "
                       "iterables or an array, and not %s" %
                       type(data_iterables))
            raise FlowException(err_str)

        # check that all elements are iterable
        for i, iterable in enumerate(data_iterables):
            if (iterable is not None) and (not hasattr(iterable, '__iter__')):
                err = ("Element number %d in the data_iterables"
                       " list is not an iterable." % i)
                raise FlowException(err)

        # check that the number of data_iterables is correct
        if len(data_iterables) != len(flow):
            err_str = ("%d data iterables specified,"
                       " %d needed" % (len(data_iterables), len(flow)))
            raise FlowException(err_str)

        return data_iterables

    def _close_last_node(self):
        if self.verbose:
            print("Close the training phase of the last node")
        try:
            self.flow[-1].stop_training()
        except mdp.TrainingFinishedException:
            pass
        except Exception as e:
            self._propagate_exception(e, len(self.flow)-1)

    def set_crash_recovery(self, state = True):
        """Set crash recovery capabilities.

        When a node raises an Exception during training, execution, or
        inverse execution that the flow is unable to handle, a FlowExceptionCR
        is raised. If crash recovery is set, a crash dump of the flow
        instance is saved for later inspection. The original exception
        can be found as the 'parent_exception' attribute of the
        FlowExceptionCR instance.

        - If 'state' = False, disable crash recovery.
        - If 'state' is a string, the crash dump is saved on a file
          with that name.
        - If 'state' = True, the crash dump is saved on a file created by
          the tempfile module.
        """
        self._crash_recovery = state

    def train(self, data_iterables):
        """Train all trainable nodes in the flow.

        'data_iterables' is a list of iterables, one for each node in the flow.
        The iterators returned by the iterables must return data arrays that
        are then used for the node training (so the data arrays are the 'x' for
        the nodes). Note that the data arrays are processed by the nodes
        which are in front of the node that gets trained, so the data dimension
        must match the input dimension of the first node.

        If a node has only a single training phase then instead of an iterable
        you can alternatively provide an iterator (including generator-type
        iterators). For nodes with multiple training phases this is not
        possible, since the iterator cannot be restarted after the first
        iteration. For more information on iterators and iterables see
        http://docs.python.org/library/stdtypes.html#iterator-types .

        In the special case that 'data_iterables' is one single array,
        it is used as the data array 'x' for all nodes and training phases.

        Instead of a data array 'x' the iterators can also return a list or
        tuple, where the first entry is 'x' and the following are args for the
        training of the node (e.g. for supervised training).
        """

        data_iterables = self._train_check_iterables(data_iterables)

        # train each Node successively
        for i in range(len(self.flow)):
            if self.verbose:
                print("Training node #%d (%s)" % (i, str(self.flow[i])))
            self._train_node(data_iterables[i], i)
            if self.verbose:
                print("Training finished")

        self._close_last_node()

    def _execute_seq(self, x, nodenr = None):
        # Filters input data 'x' through the nodes 0..'node_nr' included
        flow = self.flow
        if nodenr is None:
            nodenr = len(flow)-1
        for i in range(nodenr+1):
            try:
                x = flow[i].execute(x)
            except Exception as e:
                self._propagate_exception(e, i)
        return x

    def execute(self, iterable, nodenr = None):
        """Process the data through all nodes in the flow.

        'iterable' is an iterable or iterator (note that a list is also an
        iterable), which returns data arrays that are used as input to the flow.
        Alternatively, one can specify one data array as input.

        If 'nodenr' is specified, the flow is executed only up to
        node nr. 'nodenr'. This is equivalent to 'flow[:nodenr+1](iterable)'.
        """
        if isinstance(iterable, numx.ndarray):
            return self._execute_seq(iterable, nodenr)
        res = []
        empty_iterator = True
        for x in iterable:
            empty_iterator = False
            res.append(self._execute_seq(x, nodenr))
        if empty_iterator:
            errstr = ("The execute data iterator is empty.")
            raise FlowException(errstr)
        return numx.concatenate(res)

    def _inverse_seq(self, x):
        #Successively invert input data 'x' through all nodes backwards
        flow = self.flow
        for i in range(len(flow)-1, -1, -1):
            try:
                x = flow[i].inverse(x)
            except Exception as e:
                self._propagate_exception(e, i)
        return x

    def inverse(self, iterable):
        """Process the data through all nodes in the flow backwards
        (starting from the last node up to the first node) by calling the
        inverse function of each node. Of course, all nodes in the
        flow must be invertible.

        'iterable' is an iterable or iterator  (note that a list is also an
        iterable), which returns data arrays that are used as input to the flow.
        Alternatively, one can specify one data array as input.

        Note that this is _not_ equivalent to 'flow[::-1](iterable)',
        which also executes the flow backwards but calls the 'execute'
        function of each node."""

        if isinstance(iterable, numx.ndarray):
            return self._inverse_seq(iterable)
        res = []
        empty_iterator = True
        for x in iterable:
            empty_iterator = False
            res.append(self._inverse_seq(x))
        if empty_iterator:
            errstr = ("The inverse data iterator is empty.")
            raise FlowException(errstr)
        return numx.concatenate(res)

    def copy(self, protocol=None):
        """Return a deep copy of the flow.

        The protocol parameter should not be used.
        """
        if protocol is not None:
            _warnings.warn("protocol parameter to copy() is ignored",
                           mdp.MDPDeprecationWarning, stacklevel=2)
        return _copy.deepcopy(self)

    def save(self, filename, protocol=-1):
        """Save a pickled serialization of the flow to 'filename'.
        If 'filename' is None, return a string.

        Note: the pickled Flow is not guaranteed to be upward or
        backward compatible."""
        if filename is None:
            return _cPickle.dumps(self, protocol)
        else:
            # if protocol != 0 open the file in binary mode
            mode = 'w' if protocol == 0 else 'wb'
            with open(filename, mode) as flh:
                _cPickle.dump(self, flh, protocol)

    def __call__(self, iterable, nodenr = None):
        """Calling an instance is equivalent to call its 'execute' method."""
        return self.execute(iterable, nodenr=nodenr)

    ###### string representation

    def __str__(self):
        nodes = ', '.join([str(x) for x in self.flow])
        return '['+nodes+']'

    def __repr__(self):
        # this should look like a valid Python expression that
        # could be used to recreate an object with the same value
        # eval(repr(object)) == object
        name = type(self).__name__
        pad = len(name)+2
        sep = ',\n'+' '*pad
        nodes = sep.join([repr(x) for x in self.flow])
        return '%s([%s])' % (name, nodes)

    ###### private container methods

    def __len__(self):
        return len(self.flow)

    def _check_dimension_consistency(self, out, inp):
        """Raise ValueError when both dimensions are set and different."""
        if ((out and inp) is not None) and out != inp:
            errstr = "dimensions mismatch: %d != %d" % (out, inp)
            raise ValueError(errstr)

    def _check_nodes_consistency(self, flow = None):
        """Check the dimension consistency of a list of nodes."""
        if flow is None:
            flow = self.flow
        len_flow = len(flow)
        for i in range(1, len_flow):
            out = flow[i-1].output_dim
            inp = flow[i].input_dim
            self._check_dimension_consistency(out, inp)

    def _check_value_type_isnode(self, value):
        if not isinstance(value, mdp.Node):
            raise TypeError("flow item must be Node instance")

    def __getitem__(self, key):
        if isinstance(key, slice):
            flow_slice = self.flow[key]
            self._check_nodes_consistency(flow_slice)
            return self.__class__(flow_slice)
        else:
            return self.flow[key]

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            [self._check_value_type_isnode(item) for item in value]
        else:
            self._check_value_type_isnode(value)

        # make a copy of list
        flow_copy = list(self.flow)
        flow_copy[key] = value
        # check dimension consistency
        self._check_nodes_consistency(flow_copy)
        # if no exception was raised, accept the new sequence
        self.flow = flow_copy

    def __delitem__(self, key):
        # make a copy of list
        flow_copy = list(self.flow)
        del flow_copy[key]
        # check dimension consistency
        self._check_nodes_consistency(flow_copy)
        # if no exception was raised, accept the new sequence
        self.flow = flow_copy

    def __contains__(self, item):
        return self.flow.__contains__(item)

    def __iter__(self):
        return self.flow.__iter__()

    def __add__(self, other):
        # append other to self
        if isinstance(other, Flow):
            flow_copy = list(self.flow).__add__(other.flow)
            # check dimension consistency
            self._check_nodes_consistency(flow_copy)
            # if no exception was raised, accept the new sequence
            return self.__class__(flow_copy)
        elif isinstance(other, mdp.Node):
            flow_copy = list(self.flow)
            flow_copy.append(other)
            # check dimension consistency
            self._check_nodes_consistency(flow_copy)
            # if no exception was raised, accept the new sequence
            return self.__class__(flow_copy)
        else:
            err_str = ('can only concatenate flow or node'
                       ' (not \'%s\') to flow' % (type(other).__name__))
            raise TypeError(err_str)

    def __iadd__(self, other):
        # append other to self
        if isinstance(other, Flow):
            self.flow += other.flow
        elif isinstance(other, mdp.Node):
            self.flow.append(other)
        else:
            err_str = ('can only concatenate flow or node'
                       ' (not \'%s\') to flow' % (type(other).__name__))
            raise TypeError(err_str)
        self._check_nodes_consistency(self.flow)
        return self

    ###### public container methods

    def append(self, x):
        """flow.append(node) -- append node to flow end"""
        self[len(self):len(self)] = [x]

    def extend(self, x):
        """flow.extend(iterable) -- extend flow by appending
        elements from the iterable"""
        if not isinstance(x, Flow):
            err_str = ('can only concatenate flow'
                       ' (not \'%s\') to flow' % (type(x).__name__))
            raise TypeError(err_str)
        self[len(self):len(self)] = x

    def insert(self, i, x):
        """flow.insert(index, node) -- insert node before index"""
        self[i:i] = [x]

    def pop(self, i = -1):
        """flow.pop([index]) -> node -- remove and return node at index
        (default last)"""
        x = self[i]
        del self[i]
        return x

class CheckpointFlow(Flow):
    """Subclass of Flow class that allows user-supplied checkpoint functions
    to be executed at the end of each phase, for example to
    save the internal structures of a node for later analysis."""

    def _train_check_checkpoints(self, checkpoints):
        if not isinstance(checkpoints, list):
            checkpoints = [checkpoints]*len(self.flow)

        if len(checkpoints) != len(self.flow):
            error_str = ("%d checkpoints specified,"
                         " %d needed" % (len(checkpoints), len(self.flow)))
            raise FlowException(error_str)

        return checkpoints


    def train(self, data_iterables, checkpoints):
        """Train all trainable nodes in the flow.

        In addition to the basic behavior (see 'Node.train'), calls the
        checkpoint function 'checkpoint[i]' when the training phase of node #i
        is over.

        A checkpoint function takes as its only argument the trained node.
        If the checkpoint function returns a dictionary, its content is
        added to the instance dictionary.

        The class CheckpointFunction can be used to define user-supplied
        checkpoint functions.
        """

        data_iterables = self._train_check_iterables(data_iterables)
        checkpoints = self._train_check_checkpoints(checkpoints)

        # train each Node successively
        for i in range(len(self.flow)):
            node = self.flow[i]
            if self.verbose:
                print("Training node #%d (%s)" % (i, type(node).__name__))
            self._train_node(data_iterables[i], i)
            if (i <= len(checkpoints)) and (checkpoints[i] is not None):
                dic = checkpoints[i](node)
                if dic:
                    self.__dict__.update(dic)
            if self.verbose:
                print("Training finished")

        self._close_last_node()

class CheckpointFunction(object):
    """Base class for checkpoint functions.

    This class can be subclassed to build objects to be used as a checkpoint
    function in a CheckpointFlow. Such objects would allow to define parameters
    for the function and save informations for later use."""

    def __call__(self, node):
        """Execute the checkpoint function.

        This is the method that is going to be called at the checkpoint.
        Overwrite it to match your needs."""
        pass

class CheckpointSaveFunction(CheckpointFunction):
    """This checkpoint function saves the node in pickle format.
    The pickle dump can be done either before the training phase is finished or
    right after that.
    In this way, it is for example possible to reload it in successive sessions
    and continue the training.
    """

    def __init__(self, filename, stop_training=0, binary=1, protocol=2):
        """CheckpointSaveFunction constructor.

        'filename'      -- the name of the pickle dump file.
        'stop_training' -- if set to 0 the pickle dump is done before
                           closing the training phase
                           if set to 1 the training phase is closed and then
                           the node is dumped
        'binary'        -- sets binary mode for opening the file.
                           When using a protocol higher than 0, make sure
                           the file is opened in binary mode.
        'protocol'      -- is the 'protocol' argument for the pickle dump
                           (see Pickle documentation for details)
        """
        self.filename = filename
        self.proto = protocol
        self.stop_training = stop_training
        if binary or protocol > 0:
            self.mode = 'wb'
        else:
            self.mode = 'w'

    def __call__(self, node):
        with open(self.filename, self.mode) as fid:
            if self.stop_training:
                node.stop_training()
            _cPickle.dump(node, fid, self.proto)
